//  Theseus
//  thePOMDP.cc -- Abstract POMDP Implementation
//
//  Blai Bonet, Hector Geffner
//  Universidad Simon Bolivar, 1998, 1999, 2000, 2001, 2002

#include <iostream>

using namespace std;
//#include <theseus/theGlobal.h>
#include <theseus/thePOMDP.h>
#include <theseus/theModel.h>
#include <theseus/theBelief.h>
#include <theseus/theBeliefCache.h>
#include <theseus/theUtils.h>


///////////////////////////////////////////////////////////////////////////////
//
// POMDP Class
//

registerClass *POMDPClass::registeredElements = NULL;


POMDPClass::POMDPClass()
{
  actionCache = NULL;
  observationCache = NULL;

  beliefHash = NULL;
  theInitialBelief = NULL;
  numActions = -1;
  model = NULL;

  setParameters( 0, 1, 250, 1, 0.0 );
  reset();
}

POMDPClass::~POMDPClass( void )
{
  delete beliefHash;
  delete actionCache;
  delete observationCache;
}

void
POMDPClass::printHash( ostream& os )
{
  assert( beliefHash != NULL );
  beliefHash->print( os );
}

void 
POMDPClass::statistics( ostream& os )
{
  assert( beliefHash != NULL );
  os << "%pomdp initialBeliefValue " << getHashEntry( getTheInitialBelief() )->getValue() << endl;
  os << "%pomdp averageBranching " << ((float)branching / (float)expansions) << endl;
  os << "%pomdp actionCacheMissRatio " << (actionCache != NULL ? actionCache->missRatio() : -1.0) << endl;
  os << "%pomdp observationCacheMissRatio "<< (observationCache != NULL ? observationCache->missRatio() : -1.0) << endl;
  beliefHash->statistics( os );
}

void
POMDPClass::setHash( beliefHashClass *pBeliefHash )
{
  delete beliefHash;
  beliefHash = pBeliefHash;
}

beliefHashClass *
POMDPClass::removeHash( void )
{
  beliefHashClass *rv = beliefHash;
  beliefHash = NULL;
  return( rv );
}

void
POMDPClass::setCache( int pActionCacheSize, int pActionClusterSize,
		      int pObservationCacheSize, int pObservationClusterSize,
		      beliefClass *(*beliefConstructor)( void ) )
{
  // set belief constructor
  beliefCachedDataClass::setBeliefConstructor( beliefConstructor );
  //xxxxxx where is the destructor?

  // make action cache
  delete actionCache;
  actionCache = new beliefCacheClass( "actionCache",
				      pActionCacheSize, pActionClusterSize,
				      beliefConstructor,
				      (void *(*)()) &beliefCachedDataClass::constructor,
				      &beliefCachedDataClass::destructor );

  // make observation cache
  delete observationCache;
  observationCache = new beliefCacheClass( "observationCache",
					   pObservationCacheSize, pObservationClusterSize,
					   beliefConstructor,
					   (void *(*)()) &beliefCachedDataClass::constructor,
					   &beliefCachedDataClass::destructor );

  // keep cache sizes
  actionCacheSize = pActionCacheSize;
  actionClusterSize = pActionClusterSize;
  observationCacheSize = pObservationCacheSize;
  observationClusterSize = pObservationClusterSize;
}

void
POMDPClass::setModel( modelClass *pModel )
{
  model = pModel;
}

modelClass *
POMDPClass::getModel( void ) const
{
  return( model );
}

void
POMDPClass::setParameters( int pNumActions, int pCostPOMDP, int pMaxMoves,
			   bool pRandomTies, float pDiscountFactor )
{
  numActions = pNumActions;
  costPOMDP = pCostPOMDP;
  maxMoves = pMaxMoves;
  randomTies = pRandomTies;
  discountFactor = pDiscountFactor;
}

void
POMDPClass::setMaxMoves( int maxmoves )
{
  maxMoves = maxmoves;
}

void
POMDPClass::setRandomTies( bool randomties )
{
  randomTies = randomties;
}

void
POMDPClass::setDiscountFactor( float discountfactor )
{
  discountFactor = discountfactor;
}

float
POMDPClass::getDiscountFactor( void )
{
  return( discountFactor );
}

void
POMDPClass::reset( void )
{
  expansions = 0;
  branching = 0;
}

void
POMDPClass::cleanHash( void )
{
  beliefHash->clean();
}

POMDPClass::hashValueClass*
POMDPClass::hashValue( const beliefClass* belief )
{
  hashEntryClass *entry;
  static hashValueClass result;

  assert( belief != NULL );

  if( !(entry = beliefHash->lookup( (void*)belief )) )
    {
      result.value = belief->heuristicValue();
      result.value = false;
    }
  else
    {
      result.value = entry->getValue();
      result.solved = ((int)entry->getExtra() == 1 ? true : false);
    }
  return( &result );
}

void
POMDPClass::learnAlgorithm( resultClass& result, int extraParameter )
{
  algorithm( true, result, extraParameter );
}

void
POMDPClass::controlAlgorithm( resultClass& result, int extraParameter )
{
  algorithm( false, result, extraParameter );
}

int
POMDPClass::getNumActions( void )
{
  return( numActions );
}

bool
POMDPClass::emptyBeliefHash( void )
{
  return( beliefHash->getNumElem() == 0 );
}

void
POMDPClass::setTheInitialBelief( const beliefClass *belief )
{
  theInitialBelief = belief;
}

const beliefClass *
POMDPClass::getTheInitialBelief( void )
{
  return( theInitialBelief );
}

hashEntryClass *
POMDPClass::getHashEntry( const beliefClass *belief )
{
  return( beliefHash->lookup( (const void*) belief ) );
}

int
POMDPClass::getBestAction( const beliefClass *belief )
{
  assert( 0 );
  return( -1 );
}

const beliefClass * 
POMDPClass::getBeliefAfterAction( const beliefClass *belief, int action ) const
{
  assert( 0 );
  return( NULL );
}

const beliefClass *
POMDPClass::getBeliefAfterActionObservation( const beliefClass *belief_a, int action, int observation ) const
{
  assert( 0 );
  return( NULL );
}

void
POMDPClass::initialize( void )
{
}

void
POMDPClass::finalize( void )
{
}

void
POMDPClass::initRun( int run )
{
}

void
POMDPClass::endRun( bool lastRun )
{
  if( !lastRun )
    {
      cleanHash();
      reset();
      // the action and observation caches don't need to be cleaned
    }
}


///////////////////////
//
// serialization

void
POMDPClass::signRegistration( const char *registrationId,
			      void (*fillFunction)( istream&, POMDPClass * ), 
			      POMDPClass *(*constructor)( void ) )
{
  registerClass::insertRegistration( registeredElements,
				     registrationId,
				     (void (*)( istream&, void * )) fillFunction,
				     (void *(*)( void )) constructor );
}

void
POMDPClass::cleanup( void )
{
  registerClass::cleanup( registeredElements );
}

void
POMDPClass::write( ostream& os ) const
{
  safeWrite( &costPOMDP, sizeof( int ), 1, os );
  safeWrite( &maxMoves, sizeof( int ), 1, os );
  safeWrite( &randomTies, sizeof( bool ), 1, os );
  safeWrite( &discountFactor, sizeof( float ), 1, os );
  safeWrite( &numActions, sizeof( int ), 1, os );
  safeWrite( &actionCacheSize, sizeof( int ), 1, os );
  safeWrite( &actionClusterSize, sizeof( int ), 1, os );
  safeWrite( &observationCacheSize, sizeof( int ), 1, os );
  safeWrite( &observationClusterSize, sizeof( int ), 1, os );

  //xxxx don't write the initial belief
  //xxxx theInitialBelief->write( os );

  beliefHash->write( os );
}

POMDPClass *
POMDPClass::read( istream& is )
{
  char *id;
  POMDPClass *POMDP;

  id = registerClass::getRegisteredId( is );
  POMDP = (POMDPClass *) registerClass::readElement( is, id, registeredElements );
  delete id;
  return( POMDP );
}

void
POMDPClass::fill( istream& is, POMDPClass *pomdp )
{
  safeRead( &pomdp->costPOMDP, sizeof( int ), 1, is );
  safeRead( &pomdp->maxMoves, sizeof( int ), 1, is );
  safeRead( &pomdp->randomTies, sizeof( bool ), 1, is );
  safeRead( &pomdp->discountFactor, sizeof( float ), 1, is ); 
  safeRead( &pomdp->numActions, sizeof( int ), 1, is );
  safeRead( &pomdp->actionCacheSize, sizeof( int ), 1, is );
  safeRead( &pomdp->actionClusterSize, sizeof( int ), 1, is );
  safeRead( &pomdp->observationCacheSize, sizeof( int ), 1, is );
  safeRead( &pomdp->observationClusterSize, sizeof( int ), 1, is );

  //xxxx don't read the initial belief
  //xxxx pomdp->theInitialBelief = beliefClass::read( is );

  pomdp->beliefHash = (beliefHashClass *) hashClass::read( is );

  //xxxx this can't be done here since still don't have an initial belief
  //xxxx pomdp->setCache( pomdp->actionCacheSize, pomdp->actionClusterSize,
  //xxxx                  pomdp->observationCacheSize, pomdp->observationClusterSize,
  //xxxx                  pomdp->theInitialBelief->getConstructor() );
}
