//  Theseus
//  theQmdpHeuristic.cc -- QMDP Heuristic for Beliefs
//
//  Blai Bonet, Hector Geffner
//  Universidad Simon Bolivar, 1998, 1999, 2000, 2001, 2002

#include <math.h>
#include <iostream>
#include <typeinfo>
#include <map>

using namespace std;
#include <theseus/theGlobal.h>
#include <theseus/theTopLevel.h>
#include <theseus/theQmdpHeuristic.h>
#include <theseus/theStandardPOMDP.h>
#include <theseus/theStandardModel.h>
#include <theseus/theException.h>
#include <theseus/theUtils.h>


///////////////////////////////////////////////////////////////////////////////
//
// QMDP Heuristic Class
//

QMDPHeuristicClass::QMDPHeuristicClass()
{
}

QMDPHeuristicClass::QMDPHeuristicClass( float pDiscountFactor,
					int pNumActions,
					int pNumStates,
					const standardPOMDPClass *spomdp,
					int pExtraParameter )
{
  discountFactor = pDiscountFactor;
  numActions = pNumActions;
  numStates = pNumStates;
  extraParameter = pExtraParameter;
  pomdp = spomdp;

  compute();
}

QMDPHeuristicClass::~QMDPHeuristicClass()
{
  delete[] QMDPValue;
}

void
QMDPHeuristicClass::compute()
{
  register int   state, action, noBest;
  register float number, value, sum, localError, error;
  register float *newVector, *oldVector, *tmpVector;
  map<int,map<int,float> > *transitionModel;
  map<int,map<int,float> >::const_iterator handle;
  map<int,float>::const_iterator it;

  // allocate memory
  oldVector = new float[numStates];
  newVector = new float[numStates];
  memset( oldVector, 0, numStates * sizeof( float ) );
  memset( newVector, 0, numStates * sizeof( float ) );

  transitionModel = ((standardModelClass*)pomdp->getModel())->transitionModel;

  // solve the MDP using Gauss-Seidel Value Iteration
  do
    {
      // perform a Gauss-Seidel update
      for( state = 0; state < numStates; ++state )
	if( !pomdp->inGoal( state ) )
	  {
	    // check signal
	    if( PD.signal >= 0 )
	      goto end;

	    // do job
	    noBest = 1;
	    for( action = 0; action < numActions; ++action )
	      {
		if( pomdp->applicable( state, action ) )
		  {
		    // compute value for an action 
		    sum = 0.0;
		    number = 0.0;
		    assert( (transitionModel[action].find( state ) != transitionModel[action].end()) &&
			    !transitionModel[action][state].empty() );
		    handle = transitionModel[action].find( state );
		    for( it = (*handle).second.begin(); it != (*handle).second.end(); ++it )
		      {
			if( (*it).first < state )
			  value = newVector[(*it).first];
			else
			  value = oldVector[(*it).first];

			// update value
			switch( extraParameter )
			  {
			  case STANDARD_MODEL:
			    sum += ((*it).second * value);
			    break;
			  case UNIFORM_MODEL:
			    number += 1.0;
			    sum += value;
			    break;
			  case MIN_MODEL:
			    sum = (sum == 0.0 ? value : (value < sum ? value : sum));
			    break;
			  case MAX_MODEL:
			    sum = (sum == 0.0 ? value : (value > sum ? value : sum));
			    break;
			  }
		      }
		    if( extraParameter == UNIFORM_MODEL )
		      sum = sum / number;
		    value = pomdp->cost( state, action ) + discountFactor * sum;

		    // update maximum
		    if( noBest || (value < newVector[state]) )
		      {
			noBest = 0;
			newVector[state] = value;
		      }
		  }
	      }
	  }

      // new error
      error = 0.0;
      for( state = 0; state < numStates; ++state )
	{
	  localError = fabs( newVector[state] - oldVector[state] );
	  error = (localError > error ? localError : error);
	}

      // swap vectors
      tmpVector = oldVector;
      oldVector = newVector;
      newVector = tmpVector;
    } while( error > PD.epsilon );

  // result
 end:
  delete[] newVector;
  QMDPValue = oldVector;

  // check signal
  if( PD.signal >= 0 )
    {
      int s = PD.signal;
      PD.signal = -1;
      throw signalExceptionClass( s );
    }
}

float
QMDPHeuristicClass::value( int state )
{
  return( QMDPValue[state] );
}

float
QMDPHeuristicClass::value( const beliefClass *belief )
{
  register int number;
  register float val, sum;
  const map<int,float> *m;
  map<int,float>::const_iterator it;

  sum = 0.0;
  number = 0;
  m = belief->cast();
  for( it = m->begin(); it != m->end(); ++it )
    {
      val = value( (*it).first );
      switch( extraParameter )
	{
	case STANDARD_MODEL:
	  sum += ((*it).second * val);
	  break;
	case UNIFORM_MODEL:
	  ++number;
	  sum += val;
	  break;
	case MIN_MODEL:
	  sum = (sum == 0.0 ? val : (val < sum ? val : sum));
	  break;
	case MAX_MODEL:
	  sum = (sum == 0.0 ? val : (val > sum ? val : sum));
	  break;
	}
    }
  if( extraParameter == UNIFORM_MODEL )
    sum = sum / (float)number;
  //cout << "qmdp-value for " << *belief << " is " << sum << endl;
  return( sum );
}


///////////////////////
//
// serialization

void
QMDPHeuristicClass::checkIn( void )
{
  signRegistration( typeid( QMDPHeuristicClass ).name(),
		    (void (*)( istream&, heuristicClass * )) &QMDPHeuristicClass::fill,
		    (heuristicClass *(*)( void )) &QMDPHeuristicClass::constructor );
}

void
QMDPHeuristicClass::write( ostream& os )
{
  const char *id;
  unsigned len;

  // write registrationId
  id = typeid( *this ).name();
  len = strlen( id ) + 1;
  safeWrite( &len, sizeof( unsigned ), 1, os );
  safeWrite( (void*)id, sizeof( char ), strlen( id ) + 1, os );

  // serialize base class
  heuristicClass::write( os );

  // do it!
  safeWrite( &numStates, sizeof( int ), 1, os );
  safeWrite( &numActions, sizeof( int ), 1, os );
  safeWrite( &discountFactor, sizeof( float ), 1, os );
  safeWrite( QMDPValue, sizeof( float ), numStates, os );
}

QMDPHeuristicClass *
QMDPHeuristicClass::read( istream& is )
{
  QMDPHeuristicClass *heuristic;
  
  heuristic = new QMDPHeuristicClass;
  fill( is, heuristic );

  return( heuristic );
}

QMDPHeuristicClass *
QMDPHeuristicClass::constructor( void )
{
  return( new QMDPHeuristicClass );
}

void
QMDPHeuristicClass::fill( istream& is, QMDPHeuristicClass *heuristic )
{
  heuristicClass::fill( is, heuristic );

  safeRead( &heuristic->numStates, sizeof( int ), 1, is );
  safeRead( &heuristic->numActions, sizeof( int ), 1, is );
  safeRead( &heuristic->discountFactor, sizeof( float ), 1, is );

  heuristic->QMDPValue = new float[heuristic->numStates];
  safeRead( heuristic->QMDPValue, sizeof( float ), heuristic->numStates, is );
}
