//  Theseus
//  theSimpleBelief.cc -- Simple Belief Implementation
//
//  Blai Bonet, Hector Geffner
//  Universidad Simon Bolivar, 1998, 1999, 2000, 2001, 2002

#include <iostream>
#include <typeinfo>

using namespace std;
#include <theseus/theSimpleBelief.h>
#include <theseus/theBeliefCache.h>
#include <theseus/theStandardModel.h>
#include <theseus/theHeuristic.h>
#include <theseus/theUtils.h>


#define  PACK(a,o)     (((a)<<16)|((o)&0xFFFF))


///////////////////////////////////////////////////////////////////////////////
//
// Simple Belief Class
//

// static members
standardModelClass* simpleBeliefClass::model = NULL;


simpleBeliefClass::simpleBeliefClass( const simpleBeliefClass& belief )
{
  *this = belief;
}

simpleBeliefClass::simpleBeliefClass()
{
}

simpleBeliefClass::simpleBeliefClass( int s )
{
  state = s;
}

simpleBeliefClass::~simpleBeliefClass( void )
{
}

void
simpleBeliefClass::setModel( modelClass* m )
{
  model = dynamic_cast<standardModelClass*>( m );
}

beliefClass::constructor_t
simpleBeliefClass::getConstructor( void ) const
{
  return( (beliefClass* (*)(void))&simpleBeliefClass::constructor );
}

void
simpleBeliefClass::print( ostream& os ) const
{
  os << state;
}

// Check the consistency of a belief state. Probabilities
// must add to 1, and the registered number of nozero entries
// must be ok.
bool
simpleBeliefClass::check( void ) const
{
  return( true );
}

bool
simpleBeliefClass::check( int s ) const
{
  return( state == s );
}

const map<int,float>*
simpleBeliefClass::cast( void ) const
{
  static map<int,float> m;
  m.clear();
  m[state] = 1.0;
  return( &m );
}

void
simpleBeliefClass::insert( int s, float probability )
{
  state = s;
}

unsigned
simpleBeliefClass::randomSampling( void ) const
{
  return( state );
}

void
simpleBeliefClass::checkModelAvailability( modelClass* m ) const
{
  m->checkModelFor( state );
}

void
simpleBeliefClass::nextPossibleObservations( int action, map<int,float>& result ) const
{
  result.clear();
  assert( model->transitionModel[action].find( state ) != model->transitionModel[action].end() );
  result = (*model->transitionModel[action].find( state )).second;
}

int
simpleBeliefClass::plausibleState( int action, int observation ) const
{
  return( observation );
}

beliefClass*
simpleBeliefClass::updateByA( beliefCacheClass* cache, bool useCache, int action ) const
{
  unsigned tag;
  simpleBeliefClass *bel_a; 
  beliefCachedDataClass *cachedBel;

  tag = PACK(action,0);
  cachedBel = (beliefCachedDataClass*)cache->freeSlot( this, tag );
  bel_a = (simpleBeliefClass*)cachedBel->belief;
  assert( bel_a != this );

  *bel_a = *this;
  return( bel_a );
}

beliefClass*
simpleBeliefClass::updateByAO( beliefCacheClass* cache, bool useCache, int action, int observation ) const
{
  unsigned tag;
  simpleBeliefClass *bel_ao;
  beliefCachedDataClass *cachedBel;

  tag = PACK(action,observation);
  cachedBel = (beliefCachedDataClass*)cache->freeSlot( this, tag );
  bel_ao = (simpleBeliefClass*)cachedBel->belief;
  assert( bel_ao != this );

  bel_ao->setState( observation );
  return( bel_ao );
}

beliefClass*  
simpleBeliefClass::clone( void ) const
{
  return( new simpleBeliefClass( *this ) );
}

float
simpleBeliefClass::heuristicValue( void ) const
{
  if( heuristic != NULL )
    return( heuristic->value( this ) );
  else
    return( 0.0 );
}

void
simpleBeliefClass::support( set<int> &result ) const
{
  result.clear();
  result.insert( state );
}

int
simpleBeliefClass::supportSize( void ) const
{
  return( 1 );
}

beliefClass&
simpleBeliefClass::operator=( const beliefClass& pBel )
{
  *this = (simpleBeliefClass&)pBel;
  return( *this );
}

beliefClass& 
simpleBeliefClass::operator=( const simpleBeliefClass& pBel )
{
  state = pBel.state;
  return( *this );
}

bool
simpleBeliefClass::operator==( const beliefClass& pBel ) const
{
  return( *this == (simpleBeliefClass&) pBel );
}

bool
simpleBeliefClass::operator==( const simpleBeliefClass& pBel ) const
{
  return( state == pBel.state );
}

unsigned
simpleBeliefClass::hashDataFunction( void ) const
{
  return( state );
}

void*
simpleBeliefClass::hashDataCopy( void ) const
{
  return( (void*)clone() );
}

bool
simpleBeliefClass::hashDataCompare( const void* data ) const
{
  return( !(state == ((simpleBeliefClass*)data)->state) );
}

void
simpleBeliefClass::hashDataWrite( ostream& os ) const
{
  write( os );
}

void* 
simpleBeliefClass::hashDataRead( istream& is ) const
{
  return( beliefClass::read( is ) );
}


///////////////////////
//
// serialization

void
simpleBeliefClass::checkIn( void )
{
  signRegistration( typeid( simpleBeliefClass ).name(),
		    (void (*)( istream&, beliefClass * )) &simpleBeliefClass::fill,
		    (beliefClass* (*)( void )) &simpleBeliefClass::constructor );
}

void
simpleBeliefClass::write( ostream& os ) const
{
  const char *id;
  unsigned len;

  // write registrationId
  id = typeid( *this ).name();
  len = strlen( id ) + 1;
  safeWrite( &len, sizeof( unsigned ), 1, os );
  safeWrite( (void*)id, sizeof( char ), strlen( id ) + 1, os );

  // serialize base class
  beliefClass::write( os );

  // do it!
  os << state;
}

simpleBeliefClass*
simpleBeliefClass::read( istream& is )
{
  simpleBeliefClass *bel;

  bel = new simpleBeliefClass;
  fill( is, bel );

  return( bel );
}

simpleBeliefClass*
simpleBeliefClass::constructor( void )
{
  return( new simpleBeliefClass() );
}

void
simpleBeliefClass::fill( istream& is, beliefClass* bel )
{
  fill( is, (simpleBeliefClass*)bel );
}

void
simpleBeliefClass::fill( istream& is, simpleBeliefClass* bel )
{
  beliefClass::fill( is, bel );
  is >> bel->state;
}


///////////////////////////////////////////////////////////////////////////////
//
// Simple Belief Hash Class
//

simpleBeliefHashClass::simpleBeliefHashClass( void )
{
}

simpleBeliefHashClass::simpleBeliefHashClass( const simpleBeliefHashClass& hash )
{
  *this = hash;
}

simpleBeliefHashClass::~simpleBeliefHashClass( void )
{
  clean();
}

void
simpleBeliefHashClass::quantize( bool useCache, bool saveInHash, const beliefClass* belief,
				   hashEntryClass*& quantizationResult )
{
}

void
simpleBeliefHashClass::clean( void )
{
  beliefHashClass::clean();
}

unsigned 
simpleBeliefHashClass::dataFunction( const void* data )
{
  return( (unsigned)data );
}

void*
simpleBeliefHashClass::dataCopy( const void* data )
{
  return( (void*)data );
}

bool
simpleBeliefHashClass::dataCompare( const void* newData, const void* dataInHash )
{
  return( !(newData == dataInHash) );
}

void
simpleBeliefHashClass::dataDelete( void* data )
{
}

void
simpleBeliefHashClass::dataPrint( ostream& os, const void* data )
{
  os << (unsigned)data;
} 

void
simpleBeliefHashClass::dataWrite( ostream& os, const void* data ) const
{
  os << (unsigned)data;
}

void*
simpleBeliefHashClass::dataRead( istream& is )
{
  unsigned data;
  is >> data;
  return( (void*)data );
}

void
simpleBeliefHashClass::statistics( ostream& os )
{
  hashClass::statistics( os );
}


///////////////////////
//
// serialization

void
simpleBeliefHashClass::checkIn( void )
{
  signRegistration( typeid( simpleBeliefHashClass ).name(),
		    (void (*)( istream&, hashClass* )) &simpleBeliefHashClass::fill,
		    (hashClass* (*)( void )) &simpleBeliefHashClass::constructor );
}

void
simpleBeliefHashClass::write( ostream& os ) const
{
  const char *id;
  unsigned len;

  // write registrationId
  id = typeid( *this ).name();
  len = strlen( id ) + 1;
  safeWrite( &len, sizeof( unsigned ), 1, os );
  safeWrite( (void*)id, sizeof( char ), strlen( id ) + 1, os );

  // serialize base class
  beliefHashClass::write( os );
}

simpleBeliefHashClass*
simpleBeliefHashClass::read( istream& is )
{
  simpleBeliefHashClass *result;

  result = new simpleBeliefHashClass;
  fill( is, result );
  return( result );
}

simpleBeliefHashClass*
simpleBeliefHashClass::constructor( void )
{
  return( new simpleBeliefHashClass );
}

void
simpleBeliefHashClass::fill( istream& is, simpleBeliefHashClass* hash )
{
  beliefHashClass::fill( is, hash );
}
