//  Theseus
//  theStandardModel.cc -- Standard Model Implementation
//
//  Blai Bonet, Hector Geffner
//  Universidad Simon Bolivar, 1998, 1999, 2000, 2001, 2002

#include <stdio.h>
#include <typeinfo>

using namespace std;
#include <theseus/theTopLevel.h>
#include <theseus/theStandardModel.h>
#include <theseus/theStandardBelief.h>
#include <theseus/theRtStandard.h>
#include <theseus/theRtHash.h>
#include <theseus/theUtils.h>
#include <theseus/theException.h>


extern "C" {
#include <mdp/mdp.h>
extern int gNumGoals;
extern int gGoalList[];
}


///////////////////////////////////////////////////////////////////////////////
//
// Standard Model Class
//

standardModelClass::standardModelClass()
{
  stateHash = new stateHashClass;
  invStateHash = new invStateHashClass;
  observationHash = new observationHashClass;
  invObservationHash = new invObservationHashClass;

  stateIndex = 0;
  observationIndex = 0;

  // simple models
  applicationModel = NULL;
  costModel = NULL;

  // complex models
  transitionModel = NULL;
  observationModel = NULL;
}

standardModelClass::standardModelClass( const standardModelClass& model )
{
  theInitialBelief = NULL;
}

standardModelClass::~standardModelClass( void )
{
  delete theInitialBelief;

  // simple models
  delete[] applicationModel;
  delete[] costModel;

  // complex models
  delete[] transitionModel;
  delete[] observationModel;

  delete invObservationHash;
  delete observationHash;
  delete invStateHash;
  delete stateHash;
}

void
standardModelClass::setup( problemHandleClass *handle )
{
  register stateListClass *stateList, *state, *tmp;
  register standardBeliefClass *belief;
  register hashEntryClass *entry;

  // fundamental vars
  numActions = handle->numActions;
  stateHash->setSize( handle->stateSize );
  observationHash->setSize( handle->observationSize );

  // simple models allocation
  applicationModel = new set<int>[numActions];
  costModel = new map<int,float>[numActions];

  // complex models
  transitionModel = new map<int,map<int,float> >[numActions];
  observationModel = new map<int,map<int,float> >[numActions];

  // bootstrap problem 
  stateList = (*handle->bootstrap)();

  // insert the nullObservation into the observationHash
  observationHash->insert( handle->theNullObservation, 0, (void*)observationIndex++ );

  // set initial states and initial belief
  belief = new standardBeliefClass();
  for( state = stateList; state != NULL; state = tmp )
    {
      if( state->state->valid )
	{
	  assert( stateHash->lookup( state->state ) == NULL );
	  entry = stateHash->insert( state->state, 0, (void*)stateIndex );
	  invStateHash->insert( (void*)stateIndex, 0, entry->getData() );
	  belief->insert( stateIndex, state->probability );
	  //xxxx cout << "new state " << stateIndex << endl << *state->state << endl;
	  ++stateIndex;
	}

      // next state, free resources
      tmp = state->next;
      delete state;
    }
  theInitialBelief = belief;
  numInitialStates = stateIndex;
}

void
standardModelClass::setup( char *cassandraFilename )
{
  register int i, action, state, statePrime, observation;
  register standardBeliefClass *belief;

  // parse filename with cassandra's MDP parser
  if( !readMDP( cassandraFilename ) )
    exit( -1 );

  // fundamental vars
  numActions = gNumActions;
  discountFactor = gDiscount;

  // simple models allocation
  applicationModel = new set<int>[numActions];
  costModel = new map<int,float>[numActions];

  // complex models allocation
  transitionModel = new map<int,map<int,float> >[numActions];
  observationModel = new map<int,map<int,float> >[numActions];

  // compute model:
  for( state = 0; state < gNumStates; ++state )
    {
      for( action = 0; action < numActions; ++action )
	{
	  // check signals
	  if( PD.signal >= 0 )
	    goto end;

	  // cost model: c( s, a )
	  if( (float)getEntryMatrix( Q, action, state ) > 0.0 )
	    costModel[action][state] = getEntryMatrix( Q, action, state );

	  // transition model: P( s' | s, a )
	  for( statePrime = 0; statePrime < gNumStates; ++statePrime )
	    if( (float)getEntryMatrix( P[action], state, statePrime ) > 0.0 )
	      transitionModel[action][state][statePrime] =
		(float)getEntryMatrix( P[action], state, statePrime );

	  // observation model: P( o | s, a )
	  for( observation = 0; observation < gNumObservations; ++observation )
	    if( (float)getEntryMatrix( R[action], state, observation ) > 0.0 )
	      observationModel[action][state][observation] = 
		(float)getEntryMatrix( R[action], state, observation );

	  // application model
	  if( transitionModel[action].find( state ) != transitionModel[action].end() )
	    applicationModel[action].insert( state );
	}

      // availability model
      availabilityModel.insert( state );
    }

  // goal model
  for( i = 0; i < gNumGoals; ++i )
    goalModel.insert( gGoalList[i] );

  // set the initial belief
  numInitialStates = 0;
  belief = new standardBeliefClass();
  for( state = 0; state < gNumStates; ++state )
    if( gInitialBelief[state] > 0.0 )
      {
	++numInitialStates;
	belief->insert( state, (float)gInitialBelief[state] );
      }
  theInitialBelief = belief;

  // set some model vars
  stateIndex = gNumStates;
  observationIndex = gNumObservations;

 end:
  // check for signals 
  if( PD.signal >= 0 )
    {
      int s = PD.signal;
      PD.signal = -1;
      throw signalExceptionClass( s );
    }
}

void
standardModelClass::computeWholeModelFrom( const beliefClass *belief )
{
  const map<int,float> *m;
  map<int,float>::const_iterator it;

  m = belief->cast();
  for( it = m->begin(); it != m->end(); ++it )
    computeWholeModelFrom( (*it).first );
}

void
standardModelClass::computeWholeModelFrom( int sIndex )
{
  list<int> newStates;
  list<int>::const_iterator it;

  if( availabilityModel.find( sIndex ) == availabilityModel.end() )
    {
      computeModelFor( sIndex, newStates );
      for( it = newStates.begin(); it != newStates.end(); ++it )
	computeWholeModelFrom( *it );
    }
}

void
standardModelClass::printState( int state, ostream& os, int indent ) const
{
  ((stateClass*)invStateHash->lookup( (void*)state )->getExtra())->print( os, indent );
}

void
standardModelClass::printObservation( int obs, ostream& os, int indent ) const
{
  ((observationClass*)invObservationHash->lookup( (void*)obs )->getExtra())->print( os, indent );
}

void
standardModelClass::printData( ostream& os, const char *prefix ) const
{
  os << (prefix?prefix:"%model ") << "numberActions " << numActions << endl;
  os << (prefix?prefix:"%model ") << "numberStates " << stateIndex << endl;
  os << (prefix?prefix:"%model ") << "numberObservations " << observationIndex << endl;
  os << (prefix?prefix:"%model ") << "numberInitialStates " << numInitialStates << endl;
}

void
standardModelClass::printModel( ostream &os ) const
{
  set<int>::const_iterator it1;
  map<int,float>::const_iterator it2;
  map<int,map<int,float> >::const_iterator it3;

#if 0
  // print states
  for( int state = 0; state < stateIndex; ++state )
    {
      os << "state " << state << " = " << endl;
      printState( state, os, 4 );
      os << endl;
    }
#endif

  // print application model
  for( int action = 0; action < numActions; ++action )
    {
      os << "applicationModel[" << action << "] = { ";
      for( it1 = applicationModel[action].begin(); it1 != applicationModel[action].end(); ++it1 )
	os << *it1 << " ";
      os << "}" << endl;
    }

  // print goal model
  os << "goalModel = { ";
  for( it1 = goalModel.begin(); it1 != goalModel.end(); ++it1 )
    os << *it1 << " ";
  os << "}" << endl;

  // print cost model
  for( int action = 0; action < numActions; ++action )
    {
      os << "costModel" << action << "] = { ";
      for( it2 = costModel[action].begin(); it2 != costModel[action].end(); ++it2 )
	os << "(" << (*it2).first << "," << (*it2).second << ") ";
      os << "}" << endl;
    }

  // print transition model
  for( int action = 0; action < numActions; ++action )
    {
      os << "transitionModel for action (" << action << ") ";
      (*PD.handle->printAction)( os, action );
      os << endl;
      for( int state1 = 0; state1 < stateIndex; ++state1 )
	if( (it3 = transitionModel[action].find(state1)) != transitionModel[action].end() )
	  {
	    os << state1 << ": ";
	    for( int state2 = 0; state2 < stateIndex; ++state2 )
	      if( (it2 = (*it3).second.find(state2)) != (*it3).second.end() )
		os << "(" << (*it2).first << "," << (*it2).second << ") ";
	    os << endl;
	  }
    }

  // print observation model
  for( int action = 0; action < numActions; ++action )
    {
      os << "observationModel[" << action << "]:" << endl;
      for( int state = 0; state < stateIndex; ++state )
	if( (it3 = observationModel[action].find(state)) != observationModel[action].end() )
	  {
	    os << state << ": ";
	    for( int obs = 0; obs < observationIndex; ++obs )
	      if( (it2 = (*it3).second.find(obs)) != (*it3).second.end() )
		os << "(" << (*it2).first << "," << (*it2).second << ") ";
	    os << endl;
	  }
    }
}

int
standardModelClass::newState( void )
{
  return( stateIndex++ );
}

void
standardModelClass::newTransition( int state, int action, int statePrime, float probability )
{
  applicationModel[action].insert( state );
  if( transitionModel[action][state].find( statePrime ) != transitionModel[action][state].end() )
    transitionModel[action][state][statePrime] += probability;
  else
    transitionModel[action][state][statePrime] = probability;
}

void
standardModelClass::setAvailability( int stateIndex )
{
  availabilityModel.insert( stateIndex );
}

const stateClass*
standardModelClass::getState( int index ) const
{
  return( (const stateClass*)invStateHash->lookup( (void*)index )->getExtra() );
}

void
standardModelClass::checkModelFor( int sIndex )
{
  list<int> newStates;

  if( availabilityModel.find( sIndex ) == availabilityModel.end() )
    computeModelFor( sIndex, newStates );
}

void
standardModelClass::computeModelFor( int sIndex, list<int> &newStates )
{
  register int statePrimeIndex, observation;
  register const stateClass *state;
  register stateListClass *stateList, *sListPtr;
  register hashEntryClass *entry;
  register observationListClass *oListPtr;

  // lookup state in invStateHash
  state = (stateClass*)invStateHash->lookup( (void*)sIndex )->getExtra();

  // fill goalModel
  if( state->goal() )
    goalModel.insert( sIndex );

  // expand state
  stateList = state->applyAllActions();

  // initialize cost model for applicable actions
  for( sListPtr = stateList; sListPtr->probability != -1; ++sListPtr )
    if( sListPtr->state->valid )
      costModel[sListPtr->action].insert( make_pair( sIndex, (float)0.0 ) );

  // compute transition and observation models
  for( sListPtr = stateList; sListPtr->probability != -1; ++sListPtr )
    if( sListPtr->state->valid && (sListPtr->probability > 0.0)  )
      {
	// check signals
	if( PD.signal >= 0 )
	  goto cleanup;

	// lookup statePrimeIndex
	if( !(entry = stateHash->lookup( sListPtr->state )) )
	  {
	    // we have a new state
	    statePrimeIndex = stateIndex++;
	    entry = stateHash->insert( sListPtr->state, 0, (void*)statePrimeIndex );
	    invStateHash->insert( (void*)statePrimeIndex, 0, entry->getData() );
	    newStates.push_front( statePrimeIndex );
	    //xxxx cout << "new state " << statePrimeIndex << endl; // << *sListPtr->state << endl;
	  }
	else
	  {
	    statePrimeIndex = (int) entry->getExtra();
	  }

	// insert transition into model: P( s' | s, a )
	if( transitionModel[sListPtr->action][sIndex].find( statePrimeIndex ) !=
	    transitionModel[sListPtr->action][sIndex].end() )
	  transitionModel[sListPtr->action][sIndex][statePrimeIndex] += sListPtr->probability;
	else
	  transitionModel[sListPtr->action][sIndex][statePrimeIndex] = sListPtr->probability;

	//xxxx cerr << "T(s'="<<statePrimeIndex<<"|s="<<sIndex<<",a="<<sListPtr->action<<") = ";
	//xxxx cerr << transitionModel[sListPtr->action][sIndex][statePrimeIndex] << endl;

	assert( transitionModel[sListPtr->action].find( sIndex ) != transitionModel[sListPtr->action].end() );
	assert( transitionModel[sListPtr->action][sIndex].find( statePrimeIndex ) !=
		transitionModel[sListPtr->action][sIndex].end() );

	// observation model: P( o | s, a )
	if( observationModel[sListPtr->action].find( statePrimeIndex ) == observationModel[sListPtr->action].end() )
	  {
	    if( (sListPtr->observations == NULL) || (sListPtr->observations->probability == -1) )
	      {
		// We have an action that returns no observation.
		// Insert nullObservation with 1.0 probability.
		observationModel[sListPtr->action][statePrimeIndex][0] = 1.0;
	      }
	    else
	      {
		for( oListPtr = sListPtr->observations; oListPtr->probability != -1; ++oListPtr )
		  {
		    // We need to go to the observationHash in order to get the observationIndex
		    if( !(entry = observationHash->lookup( oListPtr->observation )) )
		      {
			// we have a new observation
			observation = observationIndex++;
			entry = observationHash->insert( oListPtr->observation, 0, (void*)observation );
			invObservationHash->insert( (void*)observation, 0, entry->getData() );
			//xxxx cout << "new obs " << observation << " = ";
			//xxxx oListPtr->observation->print( cout, 0 );
			//xxxx cout << endl;
		      }
		    else
		      {
			observation = (int) entry->getExtra();
		      }

		    // insert observation into model
		    if( observationModel[sListPtr->action][statePrimeIndex].find( observation ) !=
			observationModel[sListPtr->action][statePrimeIndex].end() )
		      observationModel[sListPtr->action][statePrimeIndex][observation] += oListPtr->probability;
		    else
		      observationModel[sListPtr->action][statePrimeIndex][observation] = oListPtr->probability;

		    //xxxx cout << "O(o="<<observation<<"|s="<<statePrimeIndex<<",a="<<sListPtr->action<<") = ";
		    //xxxx cout << observationModel[sListPtr->action][statePrimeIndex][observation] << endl << endl;

		    assert( (observationModel[sListPtr->action][statePrimeIndex][observation] >= 0.0) &&
			    (observationModel[sListPtr->action][statePrimeIndex][observation] <= 1.0) );
		  }
	      }
	  }

	// cost model: c( s, a ) = \sum_{s'} c( s, a, s' ) * P( s' | s, a' )
	costModel[sListPtr->action][sIndex] += (sListPtr->probability * sListPtr->cost);
	//xxxx cout << "cost(s=" << sIndex << ",a="<< sListPtr->action << ") = " << costModel[sListPtr->action][sIndex] << endl;

	// set application model and inversion transition model
	applicationModel[sListPtr->action].insert( sIndex );
	invTransitionModel[sIndex].insert( sListPtr->action );
      }

  // cleanup
 cleanup:
  for( sListPtr = stateList; sListPtr->probability != -1; ++sListPtr )
    delete[] sListPtr->observations;
  delete[] stateList;

  // check signals 
  if( PD.signal >= 0 )
    {
      int s = PD.signal;
      PD.signal = -1;
      throw signalExceptionClass( s );
    }

  // update availability model
  availabilityModel.insert( sIndex );
}

int
standardModelClass::getInitialState( const beliefClass *belief ) const
{
  return( belief->randomSampling() );
}

float
standardModelClass::cost( int state, int action ) const
{
  assert( action < numActions );
  if( costModel[action].find( state ) == costModel[action].end() )
    return( 0.0 );
  else
    return( costModel[action][state] );
}

float
standardModelClass::cost( const beliefClass *belief, int action ) const
{
  register float sum;
  const map<int,float> *m;
  map<int,float>::const_iterator it;

  assert( belief != NULL );

  sum = 0.0;
  m = belief->cast();
  for( it = m->begin(); it != m->end(); ++it )
    sum += cost( (*it).first, action ) * (*it).second;
  return( sum );
}

bool
standardModelClass::applicable( int state, int action ) const
{
  assert( action < numActions );
  return( applicationModel[action].find( state ) != applicationModel[action].end() );
}

bool
standardModelClass::applicable( const beliefClass* belief, int action ) const
{
  const map<int,float> *m;
  map<int,float>::const_iterator it;

  assert( belief != NULL );

  // check applicability 
  m = belief->cast();
  for( it = m->begin(); it != m->end(); ++it )
    if( !applicable( (*it).first, action ) )
      return( false );
  return( true );
}

bool
standardModelClass::inGoal( int state ) const
{
  return( goalModel.find( state ) != goalModel.end() );
}

bool
standardModelClass::inGoal( const beliefClass* belief ) const
{
  assert( belief != NULL );
  if( PD.pddlProblemType && (PD.handle->otherInfo & (1<<problemHandleClass::CERTAINTY_GOAL)) )
    {
      return( belief->supportSize() == 1 );
    }
  else
    {
      const map<int,float> *m;
      map<int,float>::const_iterator it;

      m = belief->cast();
      for( it = m->begin(); it != m->end(); ++it )
	if( !inGoal( (*it).first ) )
	  return( false );
      return( true );

#if 0
      float mass = 0.0;
      m = belief->cast();
      for( it = m->begin(); it != m->end(); ++it )
	mass += (inGoal( (*it).first ) ? (*it).second : 0.0);
      return( mass > 0.99 );
#endif
    }
}

int
standardModelClass::nextState( int state, int action ) const
{
  assert( transitionModel[action].find( state ) != transitionModel[action].end() );
  return( randomSampling( transitionModel[action][state] ) );
}

int
standardModelClass::nextObservation( int state, int action ) const
{
  // if there is no observation model, return the null observation
  assert( observationModel[action].find( state ) != observationModel[action].end() );
  return( randomSampling( observationModel[action][state] ) );
}

///////////////////////
//
// serialization

void
standardModelClass::checkIn( void )
{
  signRegistration( typeid( standardModelClass ).name(),
		    (void (*)( istream&, modelClass * )) &standardModelClass::fill,
		    (modelClass* (*)( void )) &standardModelClass::constructor );
}

void
standardModelClass::write( ostream& os ) const
{
  const char *id;
  unsigned len;
  int action;

  // write registrationId
  id = typeid( standardModelClass ).name();
  len = strlen( id ) + 1;
  safeWrite( &len, sizeof( unsigned ), 1, os );
  safeWrite( (void*)id, sizeof( char ), strlen( id ) + 1, os );

  // serialize base class
  modelClass::write( os );

  // searialize basic variables
  safeWrite( &stateIndex, sizeof( int ), 1, os );
  safeWrite( &observationIndex, sizeof( int ), 1, os );

  //xxxx must write state & observation hashes
  //xxxx stateHashClass       stateHash;
  //xxxx observationHashClass observationHash;

  // serialize models
  os << availabilityModel;
  os << goalModel;
  for( action = 0; action < numActions; ++action )
    {
      os << applicationModel[action];
      os << costModel[action];
      os << transitionModel[action];
      os << observationModel[action];
    }
  os << invTransitionModel;
}

standardModelClass*
standardModelClass::read( istream& is )
{
  standardModelClass *model;

  model = new standardModelClass;
  fill( is, model );

  return( model );
}

standardModelClass*
standardModelClass::constructor( void )
{
  return( new standardModelClass );
}

void
standardModelClass::fill( istream& is, standardModelClass *model )
{
  int action;

  // fill base class
  modelClass::fill( is, model );

  // fill basic variables
  safeRead( &model->stateIndex, sizeof( int ), 1, is );
  safeRead( &model->observationIndex, sizeof( int ), 1, is );

  //xxxx must read state & observation hahshes
  //xxxx stateHashClass       stateHash;
  //xxxx observationHashClass observationHash;

  // allocate space for filling models
  model->applicationModel = new set<int>[model->numActions];
  model->costModel = new map<int,float>[model->numActions];
  model->transitionModel = new map<int,map<int,float> >[model->numActions];
  model->observationModel = new map<int,map<int,float> >[model->numActions];

  // fill models
  is >> model->availabilityModel;
  is >> model->goalModel;
  for( action = 0; action < model->numActions; ++action )
    {
      is >> model->applicationModel[action];
      is >> model->costModel[action];
      is >> model->transitionModel[action];
      is >> model->observationModel[action];
    }
  is >> model->invTransitionModel;
}
