//  Theseus
//  theStandardPOMDP.cc -- Standard POMDP Implementation
//
//  Blai Bonet, Hector Geffner
//  Universidad Simon Bolivar, 1998, 1999, 2000, 2001, 2002

#ifdef linux
#include <values.h>
#endif
#include <math.h>
#include <iostream>
#include <typeinfo>
#include <queue>
#include <deque>
#include <stack>

using namespace std;
#include <theseus/theTopLevel.h>
#include <theseus/theStandardPOMDP.h>
#include <theseus/theStandardBelief.h>
#include <theseus/theStandardModel.h>
#include <theseus/theResult.h>
#include <theseus/theException.h>
#include <theseus/theUtils.h>


// static variables
static standardPOMDPClass::QValueClass *bestQValueStaticResult = NULL;


///////////////////////////////////////////////////////////////////////////////
//
// Standard POMDP Class
//

// static members
int standardPOMDPClass::numberStates = 0;

// static data for discretization version
static hashEntryClass*  quantizationForInitialBelief = NULL;


standardPOMDPClass::standardPOMDPClass()
{
  beliefHash = NULL;
  actionCache = NULL;
  observationCache = NULL;
}

standardPOMDPClass::~standardPOMDPClass( void )
{
}

void
standardPOMDPClass::initialize( void )
{
}

void
standardPOMDPClass::finalize( void )
{
  delete bestQValueStaticResult;
  bestQValueStaticResult = NULL;
}

void
standardPOMDPClass::initRun( int run )
{
  // quantize initial belief
  beliefHash->quantize( true, true, theInitialBelief, quantizationForInitialBelief );
}

void
standardPOMDPClass::endRun( bool lastRun )
{
  if( !lastRun )
    {
      POMDPClass::endRun( lastRun );
      quantizationForInitialBelief = NULL;      
    }
}

float
standardPOMDPClass::cost( int state, int action ) const
{
  return( model->cost( state, action ) );
}

float
standardPOMDPClass::cost( const beliefClass *belief, int action ) const
{
  return( model->cost( belief, action ) );
}

bool
standardPOMDPClass::inGoal( int state ) const
{
  return( model->inGoal( state ) );
}

bool
standardPOMDPClass::inGoal( const beliefClass* belief ) const
{
  return( model->inGoal( belief ) );
}

bool
standardPOMDPClass::applicable( int state, int action ) const
{
  return( model->applicable( state, action ) );
}

bool
standardPOMDPClass::applicable( const beliefClass* belief, int action ) const
{
  return( model->applicable( belief, action ) );
}

int
standardPOMDPClass::getInitialState( const beliefClass* belief ) const
{
  return( model->getInitialState( belief ) );
}

int
standardPOMDPClass::nextState( int state, int action ) const
{
  return( model->nextState( state, action ) );
}

int
standardPOMDPClass::nextObservation( int state, int action ) const
{
  return( model->nextObservation( state, action ) );
}

void
standardPOMDPClass::setup( int pNumberStates )
{
  numberStates = pNumberStates;
}

const beliefClass*
standardPOMDPClass::getTheInitialBelief( void )
{
  hashEntryClass *quantization;

  assert( quantizationForInitialBelief != NULL );
  quantization = quantizationForInitialBelief;
  return( (beliefClass*)quantization->getData() );
}

POMDPClass::hashValueClass*
standardPOMDPClass::hashValue( const beliefClass* belief )
{
  hashEntryClass *hashEntry;
  static hashValueClass result;

  assert( belief != NULL );

  // quantization
  beliefHash->quantize( true, true, belief, hashEntry );

  // compute value
  result.value = hashEntry->getValue();
  result.solved = (bool)hashEntry->getExtra();
  return( &result );
}

int
standardPOMDPClass::getBestAction( const beliefClass *belief )
{
  int bestAction;
  QValueClass* QResult;

  QResult = bestQValue( belief, true, 0 );
  if( QResult->numTies > 0 )
    bestAction = QResult->ties[0];
  else
    bestAction = -1;
  return( bestAction );
}

const beliefClass*
standardPOMDPClass::getBeliefAfterAction( const beliefClass *belief, int action ) const
{
  if( (action >= 0) && (action < numActions) && applicable( belief, action ) )
    return( belief->updateByA( actionCache, true, action ) );
  else
    return( NULL );
}

const beliefClass*
standardPOMDPClass::getBeliefAfterActionObservation( const beliefClass *belief_a, 
						     int action, int observation ) const
{
  beliefClass *belief_ao;
  static standardBeliefClass result;
  static map<int,float> possibleObservations;

  // for discretization version
  register hashEntryClass *quantization;

  belief_a->nextPossibleObservations( action, possibleObservations );
  if( possibleObservations.find( observation ) != possibleObservations.end() )
    {
      // get belief_ao
      belief_ao = belief_a->updateByAO( observationCache, true, action, observation );

      // get resulting belief
      beliefHash->quantize( true, true, belief_ao, quantization );
      result = *((beliefClass*)quantization->getData());
      return( &result );
    }
  return( NULL );
}

void
standardPOMDPClass::algorithm( bool learning, resultClass& result, int extraParameter )
{
  register float              realCost; 
  register int                idx, bestAction, state, new_state, observation;
  register beliefClass        *belief, *belief_a, *belief_ao;
  register QValueClass*       QResult;
  register hashEntryClass     *quantization;

  // transition graph cache
  static map<const hashEntryClass*,map<pair<int,int>,hashEntryClass*>*>::const_iterator trc_it1;
  static map<pair<int,int>,hashEntryClass*>::const_iterator trc_it2;

  // stopping rule
  stack<hashEntryClass*> Stack;

  // check we have a POMDP problem
  assert( !PD.pddlProblemType ||
	  (PD.handle->problemType == problemHandleClass::PROBLEM_POMDP) ||
	  (PD.handle->problemType == problemHandleClass::PROBLEM_ND_POMDP) );

  // initialize result
  result.runType = learning;
  result.initialValue = quantizationForInitialBelief->getValue();
  result.solved = (bool)quantizationForInitialBelief->getExtra();
  result.numMoves = 0;
  result.costToGoal = 0.0;
  result.discountedCostToGoal = 0.0;
  result.startTimer();

  // set initial belief and state
  quantization = quantizationForInitialBelief;
  belief = ((beliefClass*)quantization->getData())->clone();
  state = getInitialState( belief );

  // go for it!!!
  while( (PD.signal < 0) && (result.numMoves < maxMoves) )
    {
      // check soundness of belief and simulated system state
      assert( belief->check( state ) );
      //cout << "state = " << state;
      //cout << ", belief = " << *belief;
      //cout << ", quantization = " << quantization;
      //cout << ", HashValue = " << quantization->getValue() << endl;

      // be sure we have the model for this belief
      belief->checkModelAvailability( model );

      // stopping rule
      if( learning && PD.useStopRule )
	{
	  // check for goal state (if in learning trial check also for solved)
	  assert( quantization != NULL );
	  if( inGoal( belief ) || (learning && (bool)quantization->getExtra()) )
	    {
	      result.push_back( state, -1, -1 );
	      quantization->setExtra( (void*)true );
	      goto theEnd;
	    }
	}
      else
	{
	  // check for goal state
	  if( inGoal( belief ) )
	    {
	      result.push_back( state, -1, -1 );
	      goto theEnd;
	    }
	}

      // compute the best QValues
      QResult = bestQValue( quantization, true, extraParameter );

      // greedy selection of best action
      if( QResult->numTies > 0 )
	{
	  idx = (!randomTies ? 0 : lrand48() % QResult->numTies);
	  bestAction = QResult->ties[idx];
	}
      else
	{
	  // we have a dead-end
	  if( learning )
	    {
	      quantization->updateValue( FLT_MAX );
	      if( PD.useStopRule )
		quantization->setExtra( (void*)true );
	    }
	  result.costToGoal = -1.0;
	  result.discountedCostToGoal = -1.0;
	  result.push_back( state, -1, -1 );
	  break;
	}

      // update hash value (if learning)
      if( learning )
	{
	  assert( !PD.useStopRule || !(bool)quantization->getExtra() ||
		  (fabs( (double)(quantization->getValue()-QResult->value) ) < PD.SREpsilon) );
	  quantization->updateValue( QResult->value );

	  // stopping rule
	  if( PD.useStopRule )
	    Stack.push( quantization );
	}

      // get next state and observation
      realCost = cost( state, bestAction );
      result.costToGoal += realCost;
      result.discountedCostToGoal = realCost + discountFactor * result.discountedCostToGoal;
      new_state = nextState( state, bestAction );
      observation = nextObservation( new_state, bestAction );
      result.push_back( state, bestAction, observation );
      state = new_state;

      // get belief_ao and set new belief
      if( ((trc_it1 = trCache.find(quantization)) != trCache.end()) &&
	  ((trc_it2=(*trc_it1).second->find(make_pair(bestAction,observation)))!=(*trc_it1).second->end()) )
	{
	  quantization = (*trc_it2).second;
	  *belief = *(beliefClass*)quantization->getData();
	}
      else
	{
	  belief_a = belief->updateByA( actionCache, true, bestAction );
	  belief_ao = belief_a->updateByAO( observationCache, true, bestAction, observation );
	  beliefHash->quantize( true, true, belief_ao, quantization );
	  *belief = *((beliefClass*)quantization->getData());
	}
    }

  // dead-end reached, return fail
  result.numMoves = -1;

  // this is the end, ...
 theEnd:
  result.stopTimer();
  delete belief;

  // check for abortion
  if( PD.signal >= 0 )
    {
      // cleanup
      int s = PD.signal;
      PD.signal = -1;

      if( PD.useStopRule )
	for( ; !Stack.empty(); Stack.pop() );
      throw( signalExceptionClass( s ) );
    }

  // stopping rule
  if( PD.useStopRule && learning && (result.numMoves < maxMoves) )
    {
      slist<hashEntryClass*> closed;
      slist<hashEntryClass*>::iterator it;

      while( !Stack.empty() )
	{
	  // get entry
	  hashEntryClass *entry = Stack.top();
	  Stack.pop();

	  // check if already solved
	  if( !(bool)entry->getExtra() )
	    {
	      // try labeling
	      closed.clear();
	      if( checkSolved( entry, closed, extraParameter ) )
		{
		  for( it = closed.begin(); it != closed.end(); ++it )
		    {
		      (*it)->setExtra( (void*)true );
		      *PD.outputFile << "solved belief " << *(beliefClass*)(*it)->getData();
		      *PD.outputFile << ", value = " << (*it)->getValue() << endl;
		    }
		}
	      else
		break;
	    }
	}
    }

  // clean stack
  for( ; !Stack.empty(); Stack.pop() );
}

float
standardPOMDPClass::QValue( const beliefClass* belief, int action, 
			    bool useCache, int extraParameter )
{
  float result, sum;
  hashValueClass *hValue;
  register beliefClass *belief_a, *belief_ao; 
  static map<int,float> possibleObservations;
  static map<int,float>::const_iterator it;

  if( applicable( belief, action ) )
    {
      // compute bel_a(s)
      belief_a = belief->updateByA( actionCache, useCache, action );

      // compute possible observations
      possibleObservations.clear();
      belief_a->nextPossibleObservations( action, possibleObservations );

      // check if we have observations in the model, if not, the belief doesn't change
      if( !possibleObservations.empty() )
	{
	  // make the calculation for future costs
	  sum = 0.0;
	  for( it = possibleObservations.begin(); it != possibleObservations.end(); ++it )
	    {
	      // compute bel_ao(s)
	      belief_ao = belief_a->updateByAO( observationCache, useCache, action, (*it).first );
	      hValue = hashValue( belief_ao );

	      // update sum
	      if( PD.pddlProblemType )
		{
		  switch( PD.handle->problemType )
		    {
		    case problemHandleClass::PROBLEM_POMDP:
		      sum += ((*it).second * hValue->value);
		      break;
		    case problemHandleClass::PROBLEM_ND_POMDP:
		      sum = (sum == 0.0 ? hValue->value : (hValue->value > sum ? hValue->value : sum));
		      break;
		    default:
		      throw( unsupportedModelExceptionClass( PD.handle->problemType ) );
		    }
		}
	      else
		{
		  sum += ((*it).second * hValue->value);
		}
	    }
	}
      else
	{
	  sum = hashValue( belief_a )->value;
	}
      result = cost( belief, action ) + discountFactor * sum;
    }
  else
    {
      result = -1.0;
    }
  return( result );
}

standardPOMDPClass::QValueClass*
standardPOMDPClass::bestQValue( const beliefClass* belief, bool useCache, int extraParameter )
{
  register int action;
  register bool noBest;
  register float sum, hval;
  register beliefClass *belief_a, *belief_ao; 
  static map<int,float> possibleObservations;
  static map<int,float>::const_iterator it;
  register hashEntryClass *quantization;

  // assertion and initialization
  assert( belief != NULL );
  if( bestQValueStaticResult == NULL )
    {
      bestQValueStaticResult = new QValueClass( numActions );
    }

  noBest = true;
  ++expansions;
  bestQValueStaticResult->numTies = 0;
  for( action = 0; action < numActions; ++action )
    if( applicable( belief, action ) )
      {
	// update statistics
	++branching;

	// compute bel_a(s)
	belief_a = belief->updateByA( actionCache, useCache, action );

	// compute possible observations
	possibleObservations.clear();
	belief_a->nextPossibleObservations( action, possibleObservations );
	assert( !possibleObservations.empty() );

	// make calculation for future costs
	sum = 0.0;
	for( it = possibleObservations.begin(); it != possibleObservations.end(); ++it )
	  if( (*it).second > 0.0 )
	    {
	      // compute bel_ao(s)
	      belief_ao = belief_a->updateByAO( observationCache, useCache, action, (*it).first );
	      beliefHash->quantize( true, false, belief_ao, quantization );
	      hval = quantization->getValue();

	      // update value
	      if( PD.pddlProblemType )
		{
		  switch( PD.handle->problemType )
		    {
		    case problemHandleClass::PROBLEM_POMDP:
		      sum += ((*it).second * hval);
		      break;
		    case problemHandleClass::PROBLEM_ND_POMDP:
		      sum = (sum == 0.0 ? hval : (hval > sum ? hval : sum));
		      break;
		    default:
		      throw( unsupportedModelExceptionClass( PD.handle->problemType ) );
		    }
		}
	      else
		{
		  sum += ((*it).second * hval);
		}
	    }
	sum = cost( belief, action ) + discountFactor * sum;

	// look for the best action
	if( noBest || best( sum, bestQValueStaticResult->value ) || (sum == bestQValueStaticResult->value) )
	  {
	    if( noBest || (sum != bestQValueStaticResult->value) )
	      bestQValueStaticResult->numTies = 0;
	    noBest = false;
	    bestQValueStaticResult->value = sum;
	    bestQValueStaticResult->ties[bestQValueStaticResult->numTies] = action;
	    ++bestQValueStaticResult->numTies;
	  }
      }

  // if we have a dead-end, then result->numTies is zero.
  return( bestQValueStaticResult );
}

standardPOMDPClass::QValueClass*
standardPOMDPClass::bestQValue( const hashEntryClass* qBelief, bool useCache, int extraParameter )
{
  register int action;
  register bool noBest;
  register float sum, hval;
  register beliefClass *belief, *belief_a, *belief_ao; 
  static map<int,float> possibleObservations;
  static map<int,float>::const_iterator it;
  register hashEntryClass *quantization;

  // transition graph cache
  static map<const hashEntryClass*,map<pair<int,int>,hashEntryClass*>*>::const_iterator trc_it1;
  static map<pair<int,int>,hashEntryClass*>::const_iterator trc_it2;

  // assertion and initialization
  assert( qBelief != NULL );
  if( bestQValueStaticResult == NULL )
    {
      bestQValueStaticResult = new QValueClass( numActions );
    }

  noBest = true;
  ++expansions;
  bestQValueStaticResult->numTies = 0;
  belief = (beliefClass*)qBelief->getData();
  for( action = 0; action < numActions; ++action )
    if( applicable( belief, action ) )
      {
	// update statistics
	++branching;

	// compute bel_a(s)
	belief_a = belief->updateByA( actionCache, useCache, action );

	// compute possible observations
	possibleObservations.clear();
	belief_a->nextPossibleObservations( action, possibleObservations );
	assert( !possibleObservations.empty() );

	// check if we can insert belief_a into transition cache
	if( possibleObservations.size() == 1 ) ;

	// make calculation for future costs
	sum = 0.0;
	for( it = possibleObservations.begin(); it != possibleObservations.end(); ++it )
	  if( (*it).second > 0.0 )
	    {
	      // compute bel_ao(s)
	      if( ((trc_it1 = trCache.find(qBelief)) != trCache.end()) &&
		  ((trc_it2 = (*trc_it1).second->find(make_pair(action,(*it).first))) != (*trc_it1).second->end()) )
		{
		  hval = (*trc_it2).second->getValue();
		}
	      else
		{
		  belief_ao = belief_a->updateByAO( observationCache, useCache, action, (*it).first );
		  beliefHash->quantize( true, true, belief_ao, quantization );
		  hval = quantization->getValue();

		  if( trc_it1 == trCache.end() )
		    {
		      map<pair<int,int>,hashEntryClass*> *m = new map<pair<int,int>,hashEntryClass*>;
		      m->insert( make_pair(make_pair(action,(*it).first),quantization) );
		      trCache.insert( make_pair(qBelief,m) );
		    }
		  else
		    {
		      (*trc_it1).second->insert( make_pair(make_pair(action,(*it).first),quantization) );
		    }
		}

	      // update value
	      if( PD.pddlProblemType )
		{
		  switch( PD.handle->problemType )
		    {
		    case problemHandleClass::PROBLEM_POMDP:
		      sum += ((*it).second * hval);
		      break;
		    case problemHandleClass::PROBLEM_ND_POMDP:
		      sum = (sum == 0.0 ? hval : (hval > sum ? hval : sum));
		      break;
		    default:
		      throw( unsupportedModelExceptionClass( PD.handle->problemType ) );
		    }
		}
	      else
		{
		  sum += ((*it).second * hval);
		}
	    }
	sum = cost( belief, action ) + discountFactor * sum;

	// look for the best action
	if( noBest || best( sum, bestQValueStaticResult->value ) || (sum == bestQValueStaticResult->value) )
	  {
	    if( noBest || (sum != bestQValueStaticResult->value) )
	      bestQValueStaticResult->numTies = 0;
	    noBest = false;
	    bestQValueStaticResult->value = sum;
	    bestQValueStaticResult->ties[bestQValueStaticResult->numTies] = action;
	    ++bestQValueStaticResult->numTies;
	  }
      }

  // if we have a dead-end, then result->numTies is zero.
  return( bestQValueStaticResult );
}

bool
standardPOMDPClass::checkSolved( hashEntryClass *current, slist<hashEntryClass*>& closed, 
				 int extraParameter )
{
  register int action;
  register QValueClass  *QResult;
  register beliefClass *belief, *belief_a, *belief_ao;
  register hashEntryClass *quantization;

  static map<int,float> possibleObservations;
  static deque<hashEntryClass*> open;
  static map<int,float>::const_iterator it;
  slist<hashEntryClass*>::const_iterator it2;
  deque<hashEntryClass*>::const_iterator it3;

  // transition graph cache
  static map<const hashEntryClass*,map<pair<int,int>,hashEntryClass*>*>::const_iterator trc_it1;
  static map<pair<int,int>,hashEntryClass*>::const_iterator trc_it2;

  // initialization
  closed.clear();
  open.clear();
  open.push_back( current );

  // bfs
  while( !open.empty() )
    {
      // get first from queue
      current = open.front();
      open.pop_front();
      if( (bool)current->getExtra() )
	continue;
      closed.push_front( current );

      // check epsilon condition 
      belief = (beliefClass*)current->getData();
      belief->checkModelAvailability( model );
      QResult = bestQValue( current, true, extraParameter );
      if( QResult->numTies == 0 )
	{
	  // dead-end state
	  current->updateValue( FLT_MAX );
	  current->setExtra( (void*)true );
	  return( false );
	}
      if( fabs( QResult->value - current->getValue() ) > PD.SREpsilon )
	{
	  current->updateValue( QResult->value );
	  return( false );
	}
      action = QResult->ties[0];

      // unfold control
      possibleObservations.clear();
      belief_a = belief->updateByA( actionCache, true, action );
      belief_a->nextPossibleObservations( action, possibleObservations );
      if( !possibleObservations.empty() )
	{
	  for( it = possibleObservations.begin(); it != possibleObservations.end(); ++it )
	    if( (*it).second > 0.0 )
	      {
		// compute next belief_ao
		if( ((trc_it1 = trCache.find(current)) != trCache.end()) &&
		    ((trc_it2 = (*trc_it1).second->find(make_pair(action,(*it).first))) != (*trc_it1).second->end()) )
		  {
		    quantization = (*trc_it2).second;
		    if( inGoal( (beliefClass*)quantization->getData() ) )
		      quantization = NULL;
		  }
		else
		{
		  quantization = NULL;
		  belief_ao = belief_a->updateByAO( observationCache, true, action, (*it).first );
		  if( !inGoal( belief_ao ) )
		    beliefHash->quantize( true, true, belief_ao, quantization );
		}

		if( quantization )
		  {
		    for( it2 = closed.begin(); (it2 != closed.end()) && ((*it2) != quantization); ++it2 );
		    if( it2 == closed.end() )
		      {
			for( it3 = open.begin(); (it3 != open.end()) && ((*it3) != quantization); ++it3 );
			if( it3 == open.end() )
			  open.push_back( quantization );
		      }
		  }
	      }
	}
      else if( !inGoal( belief_a ) )
	{
	  beliefHash->quantize( true, true, belief_a, quantization );
	  for( it2 = closed.begin(); (it2 != closed.end()) && ((*it2) != quantization); ++it2 );
	  if( it2 == closed.end() )
	    {
	      for( it3 = open.begin(); (it3 != open.end()) && ((*it3) != quantization); ++it3 );
	      if( it3 == open.end() )
		open.push_back( quantization );
	    }
	}
    }
  return( true );
}


///////////////////////
//
// serialization

void
standardPOMDPClass::checkIn( void )
{
  signRegistration( typeid( standardPOMDPClass ).name(),
		    (void (*)( istream&, POMDPClass * )) &standardPOMDPClass::fill,
		    (POMDPClass *(*)( void )) &standardPOMDPClass::constructor );
}

void
standardPOMDPClass::write( ostream& os ) const
{
  const char *id;
  unsigned len;

  // write registrationId
  id = typeid( *this ).name();
  len = strlen( id ) + 1;
  safeWrite( &len, sizeof( unsigned ), 1, os );
  safeWrite( (void*)id, sizeof( char ), strlen( id ) + 1, os );

  // serialize base class
  POMDPClass::write( os );
}

standardPOMDPClass *
standardPOMDPClass::read( istream& is )
{
  standardPOMDPClass *pomdp;
  
  pomdp = new standardPOMDPClass;
  fill( is, pomdp );

  return( pomdp );
}

standardPOMDPClass *
standardPOMDPClass::constructor( void )
{
  return( new standardPOMDPClass );
}

void
standardPOMDPClass::fill( istream& is, standardPOMDPClass *pomdp )
{
  POMDPClass::fill( is, pomdp );
}
