//  Theseus
//  theTopLevel.cc -- Top Level Routines
//
//  Blai Bonet, Hector Geffner
//  Universidad Simon Bolivar, 1998, 1999, 2000, 2001, 2002

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <sys/time.h>
#include <assert.h>
#include <errno.h>

#include <iostream>
#include <fstream>

using namespace std;
#include <theseus/theGlobal.h>
#include <theseus/theTopLevel.h>
#include <theseus/theRtStandard.h>
#include <theseus/theRtRegister.h>
#include <theseus/theStandardPOMDP.h>
#include <theseus/theMDPPOMDP.h>
#include <theseus/thePlanningPOMDP.h>
#include <theseus/theSearchPOMDP.h>
#include <theseus/theStandardBelief.h>
#include <theseus/theSimpleBelief.h>
#include <theseus/thePlanningBelief.h>
#include <theseus/theSearchBelief.h>
#include <theseus/theStandardModel.h>
#include <theseus/thePlanningModel.h>
#include <theseus/theSearchModel.h>
#include <theseus/theQmdpHeuristic.h>
#include <theseus/theLookAheadHeuristic.h>
#include <theseus/theQuantization.h>
#include <theseus/theException.h>
#include <theseus/theUtils.h>


#define  MAX_PIMS          128


// software revision/version (global data)
problemClass          PD;


void
problemClass::clean( int level )
{
  switch( level )
    {
    case CLEAN_OBJECT:
      delete heuristic;
      delete belief;
      delete model;
      if( POMDP )
	{
	  delete POMDP->removeHash();
	}
      delete POMDP;
      quantizationClass::finalize();
      heuristic = NULL;
      belief = NULL;
      model = NULL;
      POMDP = NULL;
      break;

    case CLEAN_CORE:
      if( POMDP )
	{
	  delete POMDP->removeHash();
	}
      delete POMDP;
      POMDP = NULL;
      break;

    case CLEAN_HASH:
      if( POMDP )
	POMDP->cleanHash();
      break;
    }
}


static char *
readArgument( char *str, char *ctx )
{
  char *result;

  if( str == NULL )
    {
      cerr << "Error: argument expected after \"" << ctx << "\"" << endl;
      exit( -1 );
    }
  else
    {
      result = str;
    }
  return( result );
}


void
parseArguments( int argc, char **argv, void (*helpFunction)() )
{
  int pimSet, k;
  short unsigned seedArray[3];
  struct timeval tp;
  char *ctx;

  // program name
  PD.programName = argv[0];

  // set default values
  PD.softwareRevision = REVISION;
  PD.version = problemClass::VERSION_GRID;
  PD.useStopRule = false;
  PD.SREpsilon = 0.0;
  PD.epsilon = 0.000001;
  PD.signal = -1;

  pimSet = -1;
  PD.runs = 1;
  PD.pimSteps[0] = 1;
  PD.pimSteps[1] = -1;
  memset( PD.learningTrials, 0, MAX_PIMS * sizeof( int ) );
  memset( PD.controlTrials, 0, MAX_PIMS * sizeof( int ) );

  PD.cutoff = (int)(((unsigned)-1)>>1);
  PD.discretizationLevels = -1;
  PD.extraParameter = 0;
  PD.discountFactor = 1.0;
  PD.MDPdiscountFactor = 1.0;

  PD.incrementalModel = 0;
  PD.zeroHeuristic = false;
  PD.randomSeed = -1;
  PD.randomTies = true;
  PD.pddlProblemType = true;
  PD.coreLevel = 0;
  PD.loadType = 0;

  PD.outputLevel = 0;
  PD.outputFile = &cout;
  PD.outputPrefix = NULL;
  PD.coreFilename = NULL;

  // caches default sizes
  PD.cacheSize[0] = 16;
  PD.clusterSize[0] = 1;
  PD.cacheSize[1] = 16;
  PD.clusterSize[1] = 1;
  PD.cacheSize[2] = 16;
  PD.clusterSize[2] = 1;

  // parse options
  ++argv;
  while( (argc > 1) && (**argv == '-') )
    {
      ctx = *argv;
      switch( (*argv)[1] )
	{
	  // trial control
	case 'c':
	  --argc;
	  if( pimSet < 0 )
	    {
	      cerr << "Error: first specify a pim set with -p" << endl;
	      exit( -1 );
	    }
	  PD.controlTrials[pimSet] = atoi( readArgument( *++argv, ctx ) );
	  break;
	case 'l':
	  --argc;
	  if( pimSet < 0 )
	    {
	      cerr << "Error: first specify a pim set with -p" << endl;
	      exit( -1 );
	    }
	  PD.learningTrials[pimSet] = atoi( readArgument( *++argv, ctx ) );
	  break;
	case 'p':
	  --argc;
	  if( ++pimSet == 128 )
	    {
	      cerr << "Error: max pim sets reached" << endl;
	      exit( -1 );
	    }
	  PD.pimSteps[pimSet] = atoi( readArgument( *++argv, ctx ) );
	  PD.pimSteps[pimSet + 1] = -1;
	  break;
	case 'r':
	  --argc;
	  PD.runs = atoi( readArgument( *++argv, ctx ) );
	  break;

	  // general parameters
	case 'C':
	  --argc;
	  PD.cutoff = atoi( readArgument( *++argv, ctx ) );
	  break;
	case 'D':
	  --argc;
	  PD.discretizationLevels = atoi( readArgument( *++argv, ctx ) );
	  break;
	case 'g':
	  --argc;
	  PD.MDPdiscountFactor = (float) atof( readArgument( *++argv, ctx ) );
	  break;
	case 'G':
	  --argc;
	  PD.discountFactor = (float) atof( readArgument( *++argv, ctx ) );
	  break;
	case 'S':
	  --argc;
	  PD.randomSeed = atoi( readArgument( *++argv, ctx ) );
	  break;
	case 'O':
	  --argc;
	  PD.outputLevel = atoi( readArgument( *++argv, ctx ) );
	  break;
	case 'P':
	  --argc;
	  PD.outputPrefix = readArgument( *++argv, ctx );
	  break;
	case 'E':
	  --argc;
	  PD.extraParameter = atoi( readArgument( *++argv, ctx ) );
	  break;

	  // features
	case 'B':
	  PD.randomTies = false;
	  break;
	case 'I':
	  PD.incrementalModel = 1;
	  PD.zeroHeuristic = true;
	  break;
	case 'Z':
	  PD.zeroHeuristic = true;
	  break;
	case 'T':
	  PD.pddlProblemType = false;
	  break;

	  // caching
	case 'a':
	  if( (((*argv)[2] - '1') >= 0) && (((*argv)[2] - '1') <= 2) )
	    {
	      argc -= 2;
	      k = (*argv)[2] - '1';
	      PD.cacheSize[k] = atoi( readArgument( *++argv, ctx ) );
	      PD.clusterSize[k] = atoi( readArgument( *++argv, ctx ) );
	      if( (PD.cacheSize[k] < 0) || (PD.clusterSize[k] < 1) ||
		  (PD.cacheSize[k] % PD.clusterSize[k] != 0) )
		{
		  cerr << "Error: bad cache parameters: " << PD.cacheSize[k];
		  cerr << " " << PD.clusterSize[k] << endl;
		  exit( -1 );
		}
	    }
	  else
	    {
	      cerr << "Error: bad cache option." << endl;
	      exit( -1 );
	    }
	  break;

	  // core images
	case 'u':
	  if( (*argv)[2] == 'c' )
	    {
	      --argc;
	      PD.loadType = 1;
	      PD.coreFilename = readArgument( *++argv, ctx );
	    }
	  else if( (*argv)[2] == 'l' )
	    {
	      cerr << "Error: unsupported feature." << endl;
	      exit( -1 );
	    }
	  else
	    {
	      cerr << "Error: bad use core option." << endl;
	      exit( -1 );
	    }
	  break;
	case 'w':
	  if( (*argv)[2] == 'c' )
	    {
	      --argc;
	      PD.coreLevel = atoi( readArgument( *++argv, ctx ) );
	    }
	  else if( (*argv)[2] == 'l' )
	    {
	      cerr << "Error: unsupported feature." << endl;
	      exit( -1 );
	    }
	  else
	    {
	      cerr << "Error: bad write core option." << endl;
	      exit( -1 );
	    }
	  break;

	  // version
	case 'v':
	  --argc;
	  PD.version = atoi( readArgument( *++argv, ctx ) );
	  if( (PD.version != problemClass::VERSION_GRID) &&
	      (PD.version != problemClass::VERSION_ORIGINAL) )
	    {
	      cerr << "Error: bad version option." << endl;
	      exit( -1 );
	    }
	  break;

	  // help, undefined 
	case 'h':
	  (*helpFunction)();
	  exit( 0 );
	  break;
	default:
	  cerr << "usage: " << PD.programName << " <options>* <obj-file> <problem>" << endl;
	  exit( -1 );
	  break;
	}
      --argc;
      ++argv;
    }

  // if more arguments, error
  if( argc > 2 )
    {
      cerr << "usage: " << PD.programName << " <options>* [<obj-file>]" << endl;
      exit( -1 );
    }

  // What remains in argv is the name of a problem file
  if( *argv )
    {
      cout << argc << " " << *argv << endl;
      PD.problemFile = new char[1 + strlen( *argv )];
      strcpy( PD.problemFile, *argv );
    }
  else
    {
      PD.problemFile = NULL;
    }

  // set random seed
  if( PD.randomSeed == -1 )
    {
      // generate a random seed using gettimeofday
      gettimeofday( &tp, NULL );
      PD.randomSeed = (int) tp.tv_usec;
    }

  seedArray[0] = (short unsigned) PD.randomSeed;
  seedArray[1] = (short unsigned) PD.randomSeed;
  seedArray[2] = (short unsigned) PD.randomSeed;
  srand48( (long) PD.randomSeed );
  seed48( seedArray );
}


static void
classRegistration( void )
{
  // registration of all serializable classes

  // for POMDP
  standardPOMDPClass::checkIn();
  standardBeliefHashClass::checkIn();
  standardBeliefClass::checkIn();
  standardModelClass::checkIn();
  QMDPHeuristicClass::checkIn();
  lookAheadHeuristicClass::checkIn();
  hookedBeliefHashClass::checkIn();

  // for MDP
  MDPPOMDPClass::checkIn();
  simpleBeliefClass::checkIn();
  simpleBeliefHashClass::checkIn();

  // for PLANNING
  planningPOMDPClass::checkIn();
  planningBeliefHashClass::checkIn();
  planningBeliefClass::checkIn();
  planningModelClass::checkIn();

  // for CONFORMANT (search)
  searchPOMDPClass::checkIn();
  searchBeliefHashClass::checkIn();
  searchBeliefClass::checkIn();
  searchModelClass::checkIn();
}


static void
classCleanup( void )
{
  // registration of all serializable classes
  POMDPClass::cleanup();
  beliefClass::cleanup();
  modelClass::cleanup();
  heuristicClass::cleanup();
  hashClass::cleanup();
}


void
internalInitialization( void )
{
  // register all known classes
  classRegistration();

  // initialization
  //xxxx  memoryMgmt = new memoryMgmtClass;
  setOutputFormat( *PD.outputFile );
}


void
internalFinalization( void )
{
  classCleanup();
  //xxxx  delete memoryMgmt;
}


void
setOutputFormat( ostream& os )
{
  os.setf( ios::fixed | ios::showpoint );
  os.precision( PD.precision );
}


void
printParameters( ostream& os, const char *prefix )
{
  int pimSet;

  os << prefix << "softwareRevision " << PD.softwareRevision << endl;
  os << prefix << "version " << PD.version << endl;
  os << prefix << "problem \"" << PD.problemFile << "\"" << endl;
  os << prefix << "pddlProblemType " << PD.pddlProblemType << endl;
  os << prefix << "maxsteps " << PD.cutoff << endl;
  os << prefix << "discretizationLevels " << PD.discretizationLevels << endl;
  os << prefix << "discountFactor " << PD.discountFactor << endl;
  os << prefix << "MDPdiscountFactor " << PD.MDPdiscountFactor << endl;
  os << prefix << "runs " << PD.runs << endl;

  os << prefix << "pimSteps";
  for( pimSet = 0; PD.pimSteps[pimSet] != -1; ++pimSet )
    os << " " << PD.pimSteps[pimSet];
  os << endl;

  os << prefix << "learningTrials";
  for( pimSet = 0; PD.pimSteps[pimSet] != -1; ++pimSet )
    os << " " << PD.learningTrials[pimSet];
  os << endl;

  os << prefix << "controlTrials";
  for( pimSet = 0; PD.pimSteps[pimSet] != -1; ++pimSet )
    os << " " << PD.controlTrials[pimSet];
  os << endl;

  os << prefix << "incrementalModel " << PD.incrementalModel << endl;
  os << prefix << "zeroHeuristic " << PD.zeroHeuristic << endl;
  os << prefix << "randomSeed " << PD.randomSeed << endl;
  os << prefix << "outputLevel " << PD.outputLevel << endl;
}


static int
bootstrapPLANNING( problemHandleClass *handle )
{
  unsigned long secs;
  unsigned long usecs;
  struct timeval t1, t2;

  gettimeofday( &t1, NULL );
  if( (handle->beliefHook != NULL) || (handle->modelHook != NULL) )
    {
      cerr << "Error: hook not yet supported" << endl;
      return( -1 );
    }
  else
    {
      // model creation & setup
      PD.model = new planningModelClass;
      PD.belief = new standardBeliefClass;
      PD.model->setup( handle );
      PD.model->discountFactor = 1.0;
      PD.belief->setModel( PD.model );
      PD.baseHeuristic = handle->heuristicHook;
      PD.heuristic = NULL;

      // POMDP creation & setup
      PD.POMDP = new planningPOMDPClass;
      PD.POMDP->setModel( PD.model );
      PD.POMDP->setParameters( PD.model->numActions, 1, PD.cutoff, PD.randomTies, PD.model->discountFactor );
      PD.POMDP->setTheInitialBelief( PD.model->getTheInitialBelief() );
    }

  // heuristic setup
  if( !PD.zeroHeuristic && !PD.baseHeuristic )
    {
      PD.heuristic = new lookAheadHeuristicClass( PD.POMDP, PD.baseHeuristic );
      PD.heuristic->setLookahead( PD.lookahead );
      PD.belief->setHeuristic( PD.heuristic );
    }

  // timing
  gettimeofday( &t2, NULL );
  diffTime( secs, usecs, t1, t2 );
  *PD.outputFile << "%boot setupTime " << (float)secs + (float)usecs / 1000000.0 << endl;
  return( 0 );
}


static int
bootstrapMDP( problemHandleClass *handle )
{
  unsigned long secs;
  unsigned long usecs;
  struct timeval t1, t2;

  int initialState;
  int action;
  map <int,float>::const_iterator it;
  const map <int,float> *initialBelief;

  gettimeofday( &t1, NULL );
  if( (handle->beliefHook != NULL) || (handle->modelHook != NULL) )
    {
      cerr << "Error: hook not yet supported" << endl;
      return( -1 );
    }
  else
    {
      // model creation
      PD.model = new standardModelClass;
      PD.belief = new simpleBeliefClass;
      PD.baseHeuristic = handle->heuristicHook;
      PD.heuristic = NULL;

      // POMDP creation
      PD.POMDP = new MDPPOMDPClass;
      PD.POMDP->setHash( new simpleBeliefHashClass );
    }

  // model setup
  PD.model->setup( handle );
  PD.model->discountFactor = PD.discountFactor;
  PD.belief->setModel( PD.model );

  // generate extra single initial state leading to the true initial states, 
  // setup model for the extra state
  initialState = PD.model->newState();
  action = 0;                       // the INIT action geenrated by the parser
  initialBelief = PD.model->theInitialBelief->cast();
  for( it = initialBelief->begin(); it != initialBelief->end(); ++it )
    {
      PD.model->newTransition( initialState, action, (*it).first, (*it).second );
    }
  ((standardModelClass*)PD.model)->setAvailability( initialState );

  // POMDP setup
  PD.POMDP->setTheInitialBelief( (beliefClass*)initialState );
  PD.POMDP->setCache( PD.cacheSize[0], PD.clusterSize[0],
		      PD.cacheSize[1], PD.clusterSize[1],
		      (beliefClass*(*)())&simpleBeliefClass::constructor );
  PD.POMDP->setModel( PD.model );
  PD.POMDP->setParameters( PD.model->numActions, 1, PD.cutoff, PD.randomTies, PD.model->discountFactor );

  // heuristic
  if( PD.baseHeuristic != NULL )
    {
      PD.heuristic = new lookAheadHeuristicClass( PD.POMDP, PD.baseHeuristic );
      PD.heuristic->setLookahead( PD.lookahead );
      PD.belief->setHeuristic( PD.heuristic );
    }

  // timing
  gettimeofday( &t2, NULL );
  diffTime( secs, usecs, t1, t2 );
  *PD.outputFile << "%boot setupTime " << (float)secs + (float)usecs / 1000000.0 << endl;
  return( 0 );
}


static int
bootstrapPOMDP( problemHandleClass *handle )
{
  unsigned long secs;
  unsigned long usecs;
  struct timeval t1, t2;

  gettimeofday( &t1, NULL );
  if( handle->modelHook != NULL )
    {
      // model creation
      PD.model = handle->modelHook;
      PD.belief = handle->beliefHook;
      PD.baseHeuristic = handle->heuristicHook;
      PD.heuristic = NULL;

      // POMDP creation
      PD.POMDP = new standardPOMDPClass;
      PD.POMDP->setHash( new hookedBeliefHashClass );
    }
  else if( handle->beliefHook != NULL )
    {
      cerr << "Error: hook not yet supported" << endl;
      return( -1 );
    }
  else
    {
      // model creation
      PD.model = new standardModelClass;
      PD.belief = new standardBeliefClass;
      PD.baseHeuristic = handle->heuristicHook;
      PD.heuristic = NULL;

      // POMDP creation
      PD.POMDP = new standardPOMDPClass;
      PD.POMDP->setHash( new standardBeliefHashClass );
    }

  // model and belief setup
  PD.model->setup( handle );
  PD.model->discountFactor = PD.discountFactor;
  PD.belief->setModel( PD.model );

  // POMDP setup
  PD.POMDP->setTheInitialBelief( PD.model->getTheInitialBelief() );
  PD.POMDP->setCache( PD.cacheSize[0], PD.clusterSize[0],
		      PD.cacheSize[1], PD.clusterSize[1],
		      (beliefClass*(*)())&standardBeliefClass::constructor );
  PD.POMDP->setModel( PD.model );
  PD.POMDP->setParameters( PD.model->numActions, 1, PD.cutoff, PD.randomTies, PD.model->discountFactor );

  // timing
  gettimeofday( &t2, NULL );
  diffTime( secs, usecs, t1, t2 );
  *PD.outputFile << "%boot setupTime " << (float)secs + (float)usecs / 1000000.0 << endl;

  // compute model
  if( !PD.incrementalModel )
    {
      try {
	((standardModelClass*)PD.model)->computeWholeModelFrom( PD.model->getTheInitialBelief() );
      }
      catch( exceptionClass& ) {
	PD.clean( problemClass::CLEAN_OBJECT );
	throw;
      }
    }

  // timing
  gettimeofday( &t1, NULL );
  diffTime( secs, usecs, t2, t1 );
  *PD.outputFile << "%boot computeModelTime " << (float)secs + (float)usecs / 1000000.0 << endl;

  // heuristic setup
  if( !PD.zeroHeuristic )
    {
      if( !PD.baseHeuristic )
	{
	  try {
	    PD.baseHeuristic = new QMDPHeuristicClass( PD.MDPdiscountFactor,
						       PD.model->numActions,
						       ((standardModelClass *)PD.model)->stateIndex,
						       (standardPOMDPClass*)PD.POMDP,
						       PD.extraParameter );
	  }
	  catch( exceptionClass& ) {
	    PD.clean( problemClass::CLEAN_OBJECT );
	    throw;
	  }
	}
      PD.heuristic = new lookAheadHeuristicClass( PD.POMDP, PD.baseHeuristic );
      PD.heuristic->setLookahead( PD.lookahead );
    }

  // timing
  gettimeofday( &t2, NULL );
  diffTime( secs, usecs, t1, t2 );
  *PD.outputFile << "%boot computeHeuristicTime " << (float)secs + (float)usecs / 1000000.0 << endl;

  if( !PD.zeroHeuristic && PD.heuristic )
    PD.belief->setHeuristic( PD.heuristic );

  // others setup
  standardPOMDPClass::setup( ((standardModelClass *)PD.model)->stateIndex );
  quantizationClass::initialize( PD.discretizationLevels, 
				 ((standardModelClass *)PD.model)->stateIndex );
  quantizationClass::setCache( PD.cacheSize[2], PD.clusterSize[2],
			       (beliefClass*(*)())&standardBeliefClass::constructor );

  // check singal & end
  if( PD.signal >= 0 )
    {
      int s = PD.signal;
      PD.signal = -1;
      throw( signalExceptionClass( s ) );
    }
  else
    return( 0 );
}


static int
bootstrapCONFORMANT( problemHandleClass *handle )
{
  unsigned long secs;
  unsigned long usecs;
  struct timeval t1, t2;

  gettimeofday( &t1, NULL );
  if( (handle->beliefHook != NULL) || (handle->modelHook != NULL) )
    {
      cerr << "Error: hook not yet supported" << endl;
      return( -1 );
    }
  else
    {
      // model creation
      PD.model = new searchModelClass;
      PD.belief = new searchBeliefClass;
      PD.baseHeuristic = handle->heuristicHook;
      PD.heuristic = NULL;

      // POMDP creation
      PD.POMDP = new searchPOMDPClass;
      PD.POMDP->setHash( new searchBeliefHashClass );
    }

  // model and belief setup
  PD.model->setup( handle );
  PD.model->discountFactor = PD.discountFactor;
  PD.belief->setModel( PD.model );

  // POMDP setup
  PD.POMDP->setTheInitialBelief( PD.model->getTheInitialBelief() );
  PD.POMDP->setCache( PD.cacheSize[0], PD.clusterSize[0],
		      PD.cacheSize[1], PD.clusterSize[1],
		      (beliefClass*(*)())&searchBeliefClass::constructor );
  PD.POMDP->setModel( PD.model );
  PD.POMDP->setParameters( PD.model->numActions, 1, PD.cutoff, PD.randomTies, PD.model->discountFactor );

  // timing
  gettimeofday( &t2, NULL );
  diffTime( secs, usecs, t1, t2 );
  *PD.outputFile << "%boot setupTime " << (float)secs + (float)usecs / 1000000.0 << endl;

  // compute model
  if( !PD.incrementalModel )
    {
      try {
	((standardModelClass*)PD.model)->computeWholeModelFrom( PD.model->getTheInitialBelief() );
      }
      catch( exceptionClass& ) {
	PD.clean( problemClass::CLEAN_OBJECT );
	throw;
      }
    }

  // timing
  gettimeofday( &t1, NULL );
  diffTime( secs, usecs, t2, t1 );
  *PD.outputFile << "%boot computeModelTime " << (float)secs + (float)usecs / 1000000.0 << endl;

  // heuristic setup
  if( !PD.zeroHeuristic )
    {
      if( !PD.baseHeuristic )
	{
	  try {
	    PD.baseHeuristic = new QMDPHeuristicClass( PD.MDPdiscountFactor,
						       PD.model->numActions,
						       ((standardModelClass *)PD.model)->stateIndex,
						       (standardPOMDPClass*)PD.POMDP,
						       MAX_MODEL );
	  }
	  catch( exceptionClass& ) {
	    PD.clean( problemClass::CLEAN_OBJECT );
	    throw;
	  }
	}
      PD.heuristic = new lookAheadHeuristicClass( PD.POMDP, PD.baseHeuristic );
      PD.heuristic->setLookahead( PD.lookahead );
    }

  // timing
  gettimeofday( &t2, NULL );
  diffTime( secs, usecs, t1, t2 );
  *PD.outputFile << "%boot computeHeuristicTime " << (float)secs + (float)usecs / 1000000.0 << endl;

  if( !PD.zeroHeuristic && PD.heuristic )
    PD.belief->setHeuristic( PD.heuristic );

  // others setup
  searchPOMDPClass::setup( ((standardModelClass *)PD.model)->stateIndex );

  // check singal & end
  if( PD.signal >= 0 )
    {
      int s = PD.signal;
      PD.signal = -1;
      throw( signalExceptionClass( s ) );
    }
  else
    return( 0 );
}


static int
bootstrapCassandra( void )
{
  unsigned long secs;
  unsigned long usecs;
  struct timeval t1, t2;

  // model creation: use setup for cassandra's format
  gettimeofday( &t1, NULL );
  PD.model = new standardModelClass;
  PD.belief = new standardBeliefClass;
  ((standardModelClass *)PD.model)->setup( PD.problemFile );
  PD.discountFactor = PD.model->discountFactor;
  PD.belief->setModel( PD.model );

  // POMDP creation & setup
  PD.POMDP = new standardPOMDPClass;
  PD.POMDP->setTheInitialBelief( PD.model->getTheInitialBelief() );
  PD.POMDP->setHash( new standardBeliefHashClass );
  PD.POMDP->setCache( PD.cacheSize[0], PD.clusterSize[0],
		      PD.cacheSize[1], PD.clusterSize[1],
		      (beliefClass*(*)())&standardBeliefClass::constructor );
  PD.POMDP->setModel( PD.model );
  PD.POMDP->setParameters( PD.model->numActions, 1, PD.cutoff, PD.randomTies, PD.model->discountFactor );

  // timing
  gettimeofday( &t2, NULL );
  diffTime( secs, usecs, t1, t2 );
  *PD.outputFile << "%boot setupTime " << (float)secs + (float)usecs / 1000000.0 << endl;

  // heuristic setup
  if( !PD.zeroHeuristic )
    {
      PD.baseHeuristic = new QMDPHeuristicClass( PD.MDPdiscountFactor,
						 PD.model->numActions,
						 ((standardModelClass *)PD.model)->stateIndex,
						 (standardPOMDPClass*)PD.POMDP,
						 PD.extraParameter );
      PD.heuristic = new lookAheadHeuristicClass( PD.POMDP, PD.baseHeuristic );
      PD.heuristic->setLookahead( PD.lookahead );
      PD.belief->setHeuristic( PD.heuristic );
    }

  // timing
  gettimeofday( &t1, NULL );
  diffTime( secs, usecs, t2, t1 );
  *PD.outputFile << "%boot computeHeuristicTime " << (float)secs + (float)usecs / 1000000.0 << endl;

  // others setup
  standardPOMDPClass::setup( ((standardModelClass *)PD.model)->stateIndex );
  quantizationClass::initialize( PD.discretizationLevels, 
				 ((standardModelClass *)PD.model)->stateIndex );
  quantizationClass::setCache( PD.cacheSize[2], PD.clusterSize[2],
			       (beliefClass*(*)())&standardBeliefClass::constructor );

  return( 0 );
}


int
bootstrapProblem( const char *workingDir, const char *entryPoint )
{
  int rv;

  PD.clean( problemClass::CLEAN_OBJECT );
  if( PD.pddlProblemType )
    {
      // get problem handle and initialize problem
      if( !(PD.handle = getHandle( PD.problemFile, workingDir, entryPoint )) )
	return( -1 );
      (*PD.handle->initializeFunction)();

      // bootstrap specific problem
      switch( PD.handle->problemType )
	{
	case problemHandleClass::PROBLEM_PLANNING:
	  planningBeliefClass::initialize( PD.handle );
	  rv = bootstrapPLANNING( PD.handle );
	  break;
	case problemHandleClass::PROBLEM_MDP:
	case problemHandleClass::PROBLEM_ND_MDP:
	  rv = bootstrapMDP( PD.handle );
	  break;
	case problemHandleClass::PROBLEM_POMDP:
	case problemHandleClass::PROBLEM_ND_POMDP:
	  rv = bootstrapPOMDP( PD.handle );
	  break;
	case problemHandleClass::PROBLEM_CONFORMANT:
	  rv = bootstrapCONFORMANT( PD.handle );
	  break;
	default:
	  cerr << endl << "Error: undefined problem model" << endl;
	  return( -1 );
	}
    }
  else
    {
      rv = bootstrapCassandra();
    }
  return( rv );
}



void
writeCore( ostream& os )
{
  beliefHashClass *beliefHash;
  static char buff[1024];

  //xxxx In order to implement for MDP need to serialize the model; i.e.
  //xxxx the stateHash, invStateHash, ... For that need to generate code
  //xxxx for registering at bootstrap the specific state/obs classes.

  if( (PD.handle->problemType == problemHandleClass::PROBLEM_POMDP) ||
      (PD.handle->problemType == problemHandleClass::PROBLEM_ND_POMDP) )
    //      (PD.handle->problemType == problemHandleClass::PROBLEM_MDP) ||
    //      (PD.handle->problemType == problemHandleClass::PROBLEM_ND_MDP) )
    {
      // write problemType and problemFile
      strcpy( buff, basename( PD.problemFile ) );
      safeWrite( buff, sizeof( char ), 1024, os );

      // write POMDP
      beliefHash = PD.POMDP->removeHash();
      beliefHash->write( os );
      PD.POMDP->setHash( beliefHash );
    }
  else
    {
      cerr << "Error: core generation is only supported for POMDP models." << endl;
    }
}


int
readCore( const char *coreFilename )
{
  beliefHashClass *beliefHash;
  static char buff[1024];

  // xxx: only supported for MDP/POMDP problems (yet?)

  // open core file
  ifstream coreFile( coreFilename, ios::in );
  if( !coreFile )
    {
      *PD.outputFile << coreFilename << ": " << strerror( errno ) << endl;
      cerr << coreFilename << ": " << strerror( errno ) << endl;
      return( -1 );
    }

  // read problemFile
  safeRead( buff, sizeof( char ), 1024, coreFile );

  // check that core file correspond to problem
  if( strcmp( buff, basename( PD.problemFile ) ) )
    {
      cerr << "Error: core does not correspond to problem \"" << basename( PD.problemFile ) << "\"" << endl;
      return( -1 );
    }

  // read hash table from core
  delete PD.POMDP->removeHash();
  beliefHash = (beliefHashClass *)hashClass::read( coreFile );
  PD.POMDP->setHash( beliefHash );

  // close coreFile
  coreFile.close();
  return( 0 );
}


static void
generateCore( int run, int pim, int trial, int pimSet )
{
  char *tmpStr;
  const char *prefix;
  static char buff[4096];

  // generate prefix
  tmpStr = strdup( basename( PD.problemFile ) );
  prefix = (PD.outputPrefix != NULL ? PD.outputPrefix : strtok( tmpStr, "." ));

  switch( PD.coreLevel )
    {
    case 1:
      if( ((pim == PD.pimSteps[pimSet] - 1) && (trial == PD.learningTrials[pimSet] - 1)) ||
	  (PD.useStopRule && (bool)PD.POMDP->getHashEntry( PD.POMDP->getTheInitialBelief() )->getExtra()) )
	{
	  // open core file
	  sprintf( buff, "%s.core.run=%d", prefix, run + 1 );
	  ofstream coreFile( buff, ios::out /*| ios::noreplace*/ );
	  if( !coreFile )
	    {
	      *PD.outputFile << buff << ": " << strerror( errno ) << endl;
	      cerr << buff << ": " << strerror( errno ) << endl;
	    }
	  else
	    {
	      // write core and close it
	      writeCore( coreFile );
	      coreFile.close();
	    }
	}
      break;
    case 2:
      if( trial == PD.learningTrials[pimSet] - 1 )
	{
	  // open core file
	  sprintf( buff, "%s.core.run=%d.pim=%d", prefix, run + 1, pim + 1 );
	  ofstream coreFile( buff, ios::out /*| ios::noreplace*/ );
	  if( !coreFile )
	    {
	      *PD.outputFile << buff << ": " << strerror( errno ) << endl;
	      cerr << buff << ": " << strerror( errno ) << endl;
	    }
	  else
	    {
	      // write core and close it
	      writeCore( coreFile );
	      coreFile.close();
	    }
	}
      break;
    case 3:
      // open core file
      sprintf( buff, "%s.core.run=%d.pim=%d.trial=%d", prefix, run + 1, pim + 1, trial + 1 );
      ofstream coreFile( buff, ios::out /*| ios::noreplace*/ );
      if( !coreFile )
	{
	  *PD.outputFile << buff << ": " << strerror( errno ) << endl;
	  cerr << buff << ": " << strerror( errno ) << endl;
	}
      else
	{
	  // write core and close it
	  writeCore( coreFile );
	  coreFile.close();
	}
      break;
    }
  free( tmpStr );
}


static void
solvePddlProblem( void )
{
  int run, pimSet, pim, trial;
  resultClass *result;

  // build result
  switch( PD.handle->problemType )
    {
    case problemHandleClass::PROBLEM_CONFORMANT:
      result = new searchResultClass;
      break;
    case problemHandleClass::PROBLEM_MDP:
    case problemHandleClass::PROBLEM_ND_MDP:
    case problemHandleClass::PROBLEM_POMDP:
    case problemHandleClass::PROBLEM_ND_POMDP:
      result = new standardResultClass;
      break;
    default:
      cerr << endl << "Error: unsupported problem type" << endl;
      return;
    }

  // solve it
  PD.POMDP->initialize();
  for( run = 0; run < PD.runs; ++run )
    {
      result->run = run + 1;
      PD.POMDP->initRun( result->run );

      pimSet = 0;
      while( PD.pimSteps[pimSet] != -1 )
	{
	  bool solved = false;

	  // check if problem is already solved (stopRule)
	  if( (PD.handle->problemType != problemHandleClass::PROBLEM_CONFORMANT) && PD.useStopRule )
	    {
	      switch( PD.handle->problemType )
		{
		case problemHandleClass::PROBLEM_POMDP:
		case problemHandleClass::PROBLEM_ND_POMDP:
		  if( PD.POMDP->getHashEntry( PD.POMDP->getTheInitialBelief() ) &&
		      (bool)PD.POMDP->getHashEntry( PD.POMDP->getTheInitialBelief() )->getExtra() )
		    solved = true;
		  break;
		case problemHandleClass::PROBLEM_MDP:
		case problemHandleClass::PROBLEM_ND_MDP:
		  map<int,float>::const_iterator it;
		  int  state = (int)PD.POMDP->getTheInitialBelief();
		  for( it = ((standardModelClass*)PD.model)->transitionModel[0][state].begin();
		       it != ((standardModelClass*)PD.model)->transitionModel[0][state].end();
		       ++it )
		    {
		      if( !PD.POMDP->getHashEntry( (beliefClass*)(*it).first ) ||
			  !(bool)PD.POMDP->getHashEntry( (beliefClass*)(*it).first )->getExtra() )
			break;
		    }
		  if( it == ((standardModelClass*)PD.model)->transitionModel[0][state].end() )
		    solved = true;
		  break;
		}

	      if( solved )
		break;
	    }

	  // perform pim step
	  for( pim = 0; pim < PD.pimSteps[pimSet]; ++pim  )
	    {
	      // learn trial
	      trial = 0;
	      while( (trial < PD.learningTrials[pimSet]) ||
		     ((PD.handle->problemType != problemHandleClass::PROBLEM_CONFORMANT) && PD.useStopRule) )
		{
		  try {
		    PD.POMDP->learnAlgorithm( *result, PD.extraParameter );
		  } 
		  catch( exceptionClass& ) {
		    result->clean();
		    delete result;
		    PD.POMDP->endRun( true );
		    PD.POMDP->finalize();
		    throw;
		  }
		  result->print( *PD.outputFile, PD.outputLevel, PD.handle );
		  result->clean();
		  generateCore( run, pim, trial, pimSet );

		  // check termination due to stopping rule
		  if( PD.useStopRule && 
		      (bool)PD.POMDP->getHashEntry( PD.POMDP->getTheInitialBelief() )->getExtra() )
		    break;

		  // next trial
		  ++trial;
		}

	      // control trial
	      for( trial = 0; trial < PD.controlTrials[pimSet]; ++trial )
		{
		  try {
		    PD.POMDP->controlAlgorithm( *result, PD.extraParameter );
		  } 
		  catch( exceptionClass& ) {
		    result->clean();
		    delete result;
		    PD.POMDP->endRun( true );
		    PD.POMDP->finalize();
		    throw;
		  }
		  result->print( *PD.outputFile, PD.outputLevel, PD.handle );
		  result->clean();
		}
	    }

	  // next pim
	  if( (PD.handle->problemType == problemHandleClass::PROBLEM_CONFORMANT) || !PD.useStopRule )
	    ++pimSet;
	}

      // check abortion
      if( PD.signal >= 0 )
	{
	  int s = PD.signal;
	  PD.signal = -1;

	  result->clean();
	  delete result;
	  PD.POMDP->endRun( true );
	  PD.POMDP->finalize();
	  throw( signalExceptionClass( s ) );
	}

      // statistics and clean
      PD.POMDP->statistics( *PD.outputFile );

      // finish this run
      PD.POMDP->endRun( (run == PD.runs-1) );
    }
  PD.POMDP->finalize();

  // free resources
  delete result;
}


static void
solveCassandraProblem( void )
{
  int run, pimSet, pim, trial;
  resultClass *result;

  // build result
  result = new standardResultClass;

  // solve it
  PD.POMDP->initialize();
  for( run = 0; run < PD.runs; ++run )
    {
      result->run = run + 1;
      PD.POMDP->initRun( result->run );

      pimSet = 0;
      while( PD.pimSteps[pimSet] != -1 )
	{
	  bool solved = false;

	  // check if problem is already solved (stopRule)
	  if( PD.useStopRule )
	    {
	      if( PD.POMDP->getHashEntry( PD.POMDP->getTheInitialBelief() ) &&
		  (bool)PD.POMDP->getHashEntry( PD.POMDP->getTheInitialBelief() )->getExtra() )
		{
		  solved = true;
		  break;
		}
	    }

	  // perform pim step
	  for( pim = 0; pim < PD.pimSteps[pimSet]; ++pim  )
	    {
	      // learn trial
	      trial = 0;
	      while( (trial < PD.learningTrials[pimSet]) || PD.useStopRule )
		{
		  try {
		    PD.POMDP->learnAlgorithm( *result, PD.extraParameter );
		  } 
		  catch( exceptionClass& ) {
		    result->clean();
		    delete result;
		    PD.POMDP->endRun( true );
		    PD.POMDP->finalize();
		    throw;
		  }
		  result->print( *PD.outputFile, PD.outputLevel, PD.handle );
		  result->clean();
		  generateCore( run, pim, trial, pimSet );

		  // check termination due to stopping rule
		  if( PD.useStopRule && 
		      (bool)PD.POMDP->getHashEntry( PD.POMDP->getTheInitialBelief() )->getExtra() )
		    break;

		  // next trial
		  ++trial;
		}

	      // control trial
	      for( trial = 0; trial < PD.controlTrials[pimSet]; ++trial )
		{
		  try {
		    PD.POMDP->controlAlgorithm( *result, PD.extraParameter );
		  } 
		  catch( exceptionClass& ) {
		    result->clean();
		    delete result;
		    PD.POMDP->endRun( true );
		    PD.POMDP->finalize();
		    throw;
		  }
		  result->print( *PD.outputFile, PD.outputLevel, PD.handle );
		  result->clean();
		}
	    }

	  // next pim
	  if( !PD.useStopRule )
	    ++pimSet;
	}

      // check abortion
      if( PD.signal >= 0 )
	{
	  int s = PD.signal;
	  PD.signal = -1;

	  result->clean();
	  delete result;
	  PD.POMDP->endRun( true );
	  PD.POMDP->finalize();
	  throw( signalExceptionClass( s ) );
	}

      // statistics and clean
      PD.POMDP->statistics( *PD.outputFile );

      // finish this run
      PD.POMDP->endRun( (run == PD.runs-1) );
    }
  PD.POMDP->finalize();

  // free resources
  delete result;
}


void
solveProblem( void )
{
  if( PD.pddlProblemType )
    solvePddlProblem();
  else
    solveCassandraProblem();
}


///////////////////////////////////////////////////////////////////////////////
//
// Policy Generation
//


policyGraphClass::~policyGraphClass()
{
  set<policyGraphNodeClass>::const_iterator it;
  for( it = nodes.begin(); it != nodes.end(); ++it )
    delete (*it).second;
}

const policyGraphNodeClass&
policyGraphClass::retrieveNode( const policyGraphNodeClass& node )
{
  set<policyGraphNodeClass>::const_iterator it;

  for( it = nodes.begin(); it != nodes.end(); ++it )
    if( ((*it).first == node.first) && (*((*it).second) == *(node.second)) )
      {
	delete node.second;
	return( *it );
      }
  nodes.insert( node );
  return( *(nodes.find( node )) );
}

static void
expandBeliefWithAction( const beliefClass *belief_a, int action, 
			map<int,pair<float,const beliefClass*> > &result )
{
  const beliefClass *belief_ao;
  map<int,float>::const_iterator it;
  static map<int,float> possibleObservations;

  // make sure we have the model
  belief_a->checkModelAvailability( PD.model );

  // calculate possible next observations (and QValue)
  possibleObservations.clear();
  belief_a->nextPossibleObservations( action, possibleObservations );
#if 0
  cout << "belief_a = " << *belief_a << endl;
#endif
  if( !possibleObservations.empty() )
    {
      for( it = possibleObservations.begin(); it != possibleObservations.end(); ++it )
	if( (*it).second > 0.0 )
	  {
	    belief_ao = PD.POMDP->getBeliefAfterActionObservation( belief_a, action, (*it).first );
#if 0
	    cout << "belief_ao = " << *belief_ao << endl;
#endif
	    result.insert( make_pair( (*it).first, make_pair( (*it).second, (const beliefClass*)belief_ao->clone() ) ) );
	  }
    }
}

static int
getBestAction( const beliefClass *belief )
{
  bool noBest;
  int action, bestAction;
  float QValue, bestQValue;
  const beliefClass *belief_a, *belief_ao;
  map<int,float>::const_iterator it;
  static map<int,float> possibleObservations;

  // make sure we have the model
  belief->checkModelAvailability( PD.model );

  noBest = true;
  for( action = 0; action < PD.POMDP->getNumActions(); ++action )
    if( PD.POMDP->applicable( belief, action ) )
      {
	// get belief after action
	belief_a = PD.POMDP->getBeliefAfterAction( belief, action );
	belief_a->checkModelAvailability( PD.model );

	// calculate possible next observations (and QValue)
	possibleObservations.clear();
	belief_a->nextPossibleObservations( action, possibleObservations );

	// retrieve action cost
	if( (PD.handle->problemType == problemHandleClass::PROBLEM_MDP) ||
	    (PD.handle->problemType == problemHandleClass::PROBLEM_ND_MDP) )
	  QValue = PD.POMDP->cost( (beliefClass*)((simpleBeliefClass*)belief)->getState(), action );
	else
	  QValue = PD.POMDP->cost( belief, action );

	if( !possibleObservations.empty() )
	  {
	    for( it = possibleObservations.begin(); it != possibleObservations.end(); ++it )
	      if( (*it).second > 0.0 )
		{
		  belief_ao = PD.POMDP->getBeliefAfterActionObservation( belief_a, action, (*it).first );
		  if( (PD.handle->problemType == problemHandleClass::PROBLEM_MDP) ||
		      (PD.handle->problemType == problemHandleClass::PROBLEM_ND_MDP) )
		    QValue += (*it).second * PD.POMDP->hashValue( (beliefClass*)((simpleBeliefClass*)belief_ao)->getState() )->value;
		  else
		    QValue += (*it).second * PD.POMDP->hashValue( belief_ao )->value;
		}
	  }
	else
	  {
	    QValue += PD.POMDP->hashValue( belief_a )->value;
	  }

	// update best QValue
	if( noBest || (QValue < bestQValue) )
	  {
	    noBest = false;
	    bestQValue = QValue;
	    bestAction = action;
	  }
      }
  return( bestAction );
}

void
policyGraphClass::generatePolicyGraphAux( policyGraphNodeClass& node )
{
  int obs;
  const beliefClass *belief_a, *belief_ao;
  policyGraphNodeClass child_node;
  map<int,pair<float,const beliefClass*> > expansion;
  map<int,pair<float,const beliefClass*> >::const_iterator it;

#if 0
  cout << "GRAPH: insert belief " << node.second << " = " << *(node.second) << endl;
  cout << "GRAPH: with action " << node.first << endl;
#endif

  // check for goal node
  if( PD.POMDP->inGoal( node.second ) )
    {
      goals.insert( node );
      return;
    }

  // check for dead end
  if( node.first == -1 )
    return;

  // generate and expand belief_a
  belief_a = PD.POMDP->getBeliefAfterAction( node.second, node.first );
  expandBeliefWithAction( belief_a, node.first, expansion );
  assert( expansion.size() >= 1 );

  // go over children
  for( it = expansion.begin(); it != expansion.end(); ++it )
    {
      // create node
      belief_ao = (*it).second.second;
      child_node = retrieveNode( make_pair( PD.POMDP->getBestAction( belief_ao ), belief_ao ) );

      // create edge
      obs = (*it).first;
      if( edges.find( make_pair( obs, make_pair( node, child_node ) ) ) == edges.end() )
	{
	  edges.insert( make_pair( obs, make_pair( node, child_node ) ) );
#if 0
	  cout << "GRAPH: insert edge <" << node.second << "," << child_node.second << ">" << endl;
#endif
	  generatePolicyGraphAux( child_node );
#if 0
	  if( node.second == child_node.second )
	    {
	      cout << "fatal error in GRAPH generation. Please contact support." << endl;
	    }
#endif
	}
    }
}

void
policyGraphClass::generatePolicyGraph( void )
{
  int state;
  policyTableNodeClass root;
  map<int,float>::const_iterator it;

  if( PD.pddlProblemType &&
      ((PD.handle->problemType == problemHandleClass::PROBLEM_MDP) ||
       (PD.handle->problemType == problemHandleClass::PROBLEM_ND_MDP)) )
    {
      state = (int)PD.POMDP->getTheInitialBelief();
      for( it = ((standardModelClass*)PD.model)->transitionModel[0][state].begin();
	   it != ((standardModelClass*)PD.model)->transitionModel[0][state].end();
	   ++it )
	{
	  root.second = new simpleBeliefClass( (*it).first );
	  root.first = PD.POMDP->getBestAction( root.second );
	  nodes.insert( root );
	  roots.insert( root );
	  generatePolicyGraphAux( root );
	}
    }
  else
    {
      root.second = PD.POMDP->getTheInitialBelief()->clone();
      root.first = PD.POMDP->getBestAction( root.second );
      nodes.insert( root );
      roots.insert( root );
      generatePolicyGraphAux( root );
    }
}

void
policyGraphClass::outputPolicyGraph( ostream &out )
{
  int id;
  set<policyGraphEdgeClass>::const_iterator edge_it;
  set<policyGraphNodeClass>::const_iterator node_it;

  out << "Creator \"GPT " << PD.softwareRevision << "\"" << endl;
  out << "Version 1.7" << endl;
  out << "graph [ " << endl << endl;
  out << "  directed 1" << endl << endl;

  // nodes
  for( node_it = nodes.begin(), id = 0; node_it != nodes.end(); ++node_it, ++id )
    {
      out << "  node [" << endl;
      out << "    id " << id << endl;

      if( goals.find( (*node_it) ) == goals.end() )
	{
	  out << "    label \"";
	  if( PD.pddlProblemType )
	    (*PD.handle->printAction)( out, (*node_it).first );
	  else
	    out << (*node_it).first;
	  out << "\"" << endl;
	}
      else
	{
	  out << "    label \"GOAL\"" << endl;
	}

      out << "      labelAnchor \"w\"" << endl;
      out << "    graphics [" << endl;
      out << "      type \"rectangle\"" << endl;
      out << "      x " << mrand48() % 1000 << ".0" << endl;
      out << "      y " << mrand48() % 1000 << ".0" << endl;
      out << "      w 30.0" << endl;
      out << "      h 30.0" << endl;

      if( roots.find( (*node_it) ) != roots.end() )
	out << "      fill \"#00FFFF\"" << endl;
      else if( goals.find( (*node_it) ) != goals.end() )
	out << "      fill \"#FFFF00\"" << endl;
      else
	out << "      fill \"#FA1340\"" << endl;

      out << "      width 1.0" << endl;
      out << "      outline \"#000000\"" << endl;
      out << "    ]" << endl;
      out << "    LabelGraphics [" << endl;
      out << "      type \"text\"" << endl;
      out << "      fill \"#000000\"" << endl;
      out << "      anchor \"w\"" << endl;
      out << "    ]" << endl;
      out << "  ]" << endl;
    }
  out << endl;

  // edges
  for( edge_it = edges.begin(); edge_it != edges.end(); edge_it++ )
    {
      out << "  edge [" << endl;

      // find source id
      for( node_it = nodes.begin(), id = 0; node_it != nodes.end(); ++node_it, ++id )
	if( (*node_it) == (*edge_it).second.first )
	  {
	    out << "    source " << id << endl;
	    break;
	  }

      // find target id
      for( node_it = nodes.begin(), id = 0; node_it != nodes.end(); ++node_it, ++id )
	if( (*node_it) == (*edge_it).second.second )
	  {
	    out << "    target " << id << endl;
	    break;
	  }

      // edge label
      out << "    label \"obs:" << (*edge_it).first << "\"" << endl;
      out << "    graphics [" << endl;
      out << "      type \"line\"" << endl;
      out << "      arrow \"last\"" << endl;
      out << "    ]" << endl;
      out << "    labelGraphics [" << endl;
      out << "      type \"text\"" << endl;
      out << "      fill \"#000000\"" << endl;
      out << "      anchor \"e\"" << endl;
      out << "    ]" << endl;
      out << "  ]" << endl;
    }
  out << endl << "]" << endl;
}

policyTableClass::~policyTableClass()
{
  set<policyTableNodeClass>::const_iterator it;
  for( it = nodes.begin(); it != nodes.end(); ++it )
    delete (*it).second;
}

bool
policyTableClass::inserted( const policyTableNodeClass& node )
{
  set<policyTableNodeClass>::const_iterator it;

  for( it = nodes.begin(); it != nodes.end(); ++it )
    if( ((*it).first == node.first) && (*((*it).second) == *(node.second)) )
      {
	return( true );
      }
  return( false );
}

void
policyTableClass::generatePolicyTableAux( policyTableNodeClass& node )
{
  const beliefClass *belief_a, *belief_ao;
  policyTableNodeClass child_node;
  map<int,pair<float,const beliefClass*> > expansion;
  map<int,pair<float,const beliefClass*> >::const_iterator it;
  const map<int,float> *m;
  map<int,float>::const_iterator it2;

  // insert states
  m = node.second->cast();
  for( it2 = m->begin(); it2 != m->end(); ++it2 )
    states.insert( (*it2).first );

  // check for goal node
  if( PD.POMDP->inGoal( node.second ) )
    {
      goals.insert( node );
      return;
    }

  // check for dead end
  if( node.first == -1 )
    return;

  // generate and expand belief_a
  belief_a = PD.POMDP->getBeliefAfterAction( node.second, node.first );
  expandBeliefWithAction( belief_a, node.first, expansion );
  assert( expansion.size() >= 1 );

  // go over children
  for( it = expansion.begin(); it != expansion.end(); ++it )
    {
      // create new node
      belief_ao = (*it).second.second;
      child_node = make_pair( PD.POMDP->getBestAction( belief_ao ), belief_ao );

      if( !inserted( child_node ) )
	{
	  nodes.insert( child_node );
	  generatePolicyTableAux( child_node );
	}
      else
	{
	  delete child_node.second;
	}
    }
}

void
policyTableClass::generatePolicyTable( void )
{
  int state;
  policyTableNodeClass root;
  map<int,float>::const_iterator it;

  if( PD.pddlProblemType &&
      ((PD.handle->problemType == problemHandleClass::PROBLEM_MDP) ||
       (PD.handle->problemType == problemHandleClass::PROBLEM_ND_MDP)) )
    {
      state = (int)PD.POMDP->getTheInitialBelief();
      for( it = ((standardModelClass*)PD.model)->transitionModel[0][state].begin();
	   it != ((standardModelClass*)PD.model)->transitionModel[0][state].end();
	   ++it )
	{
	  root.second = new simpleBeliefClass( (*it).first );
	  root.first = PD.POMDP->getBestAction( root.second );
	  nodes.insert( root );
	  generatePolicyTableAux( root );
	}
    }
  else
    {
      root.second = PD.POMDP->getTheInitialBelief()->clone();
      root.first = PD.POMDP->getBestAction( root.second );
      nodes.insert( root );
      generatePolicyTableAux( root );
    }
}

void
policyTableClass::outputPolicyTable( ostream &out )
{
  set<int>::const_iterator state_it;
  set<policyTableNodeClass>::const_iterator node_it;

  // states
  out << "*****  STATES  *****" << endl << endl;
  for( state_it = states.begin(); state_it != states.end(); ++state_it )
    {
      out << "state-" << *state_it << ":" << endl;
      PD.model->printState( *state_it, out, 4 );
      out << endl;
    }

  // nodes
  out << "*****  POLICY  *****" << endl << endl;
  for( node_it = nodes.begin(); node_it != nodes.end(); ++node_it )
    {
      out << "(" << *(*node_it).second << ",";

      if( goals.find( (*node_it) ) == goals.end() )
	{
	  if( (*node_it).first != -1 )
	    {
	      if( PD.pddlProblemType )
		(*PD.handle->printAction)( out, (*node_it).first );
	      else
		out << (*node_it).first;
	    }
	  else
	    {
	      out << "DEAD-END";
	    }
	}
      else
	out << "GOAL";
      out << ")" << endl;
    }
  out << endl;
}
