//  Theseus
//  theTriangulation.cc -- Freudenthal Triangulation
//
//  Blai Bonet, Hector Geffner
//  Universidad Simon Bolivar, 1998, 1999, 2000, 2001, 2002

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <stl.h>
#include <iostream>

using namespace std;
#include <theseus/theTopLevel.h>

// avoid Anachronism Warning
extern "C" {
static int   pairCompare( const void* v1, const void* v2 );
}


///////////////////////////////////////////////////////////////////////////////
//
// Triangulation Cached Data Class
//

class triangulationCachedDataClass
{
public:
  // members
  map<int,float>    barycentricCoordinates;
  hashEntryClass**  simplex;

  // static members
  static int dimensions;

  // constructors
  triangulationCachedDataClass();

  // destructors
  ~triangulationCachedDataClass();

  // static methods
  static void                          setupDimensions( int dimensions );
  static triangulationCachedDataClass* constructor( void );
  static void                          destructor( void* cachedData );
};

int triangulationCachedDataClass::dimensions = 0;


triangulationCachedDataClass::triangulationCachedDataClass()
{
  simplex = new hashEntryClass*[dimensions+1];
}

triangulationCachedDataClass::~triangulationCachedDataClass()
{
  delete[] simplex;
}

void
triangulationCachedDataClass::setupDimensions( int pDimensions )
{
  dimensions = pDimensions;
}

triangulationCachedDataClass*
triangulationCachedDataClass::constructor( void )
{
  return( new triangulationCachedDataClass );
}

void
triangulationCachedDataClass::destructor( void* cachedData )
{
  delete (triangulationCachedDataClass*)cachedData;
}


///////////////////////////////////////////////////////////////////////////////
//
// Triangulation Class
//

// structure
struct pair_s
{
  int   idx;
  float val;
};

// static data
static float*                lambda = NULL;
static float*                base = NULL;
static float*                x = NULL;
static pair_s*               delta = NULL;
static float**               vec = NULL;
static standardBeliefClass*  simplexCorner = NULL;

// static members
int                          triangulationClass::dimensions = 0;
float                        triangulationClass::levels = 0.0;
beliefCacheClass*            triangulationClass::cache = NULL;


triangulationClass::triangulationClass()
{
}

triangulationClass::~triangulationClass()
{
}

void
triangulationClass::initialize( int pLevels, int pDimensions )
{
  register int i;

  // set parameters
  levels = (float)pLevels;
  dimensions = pDimensions;

  // allocate working space
  lambda = new float[dimensions];
  base = new float[dimensions];
  delta = new pair_s[dimensions];
  x = new float[dimensions];

  simplexCorner = new standardBeliefClass;
  vec = new float*[dimensions+1];
  for( i = 0; i < dimensions + 1; ++i )
    vec[i] = new float[dimensions];
}

void
triangulationClass::setCache( int cacheSize, int clusterSize,
			      beliefClass* (*beliefConstructor)( void ) )
{
  triangulationCachedDataClass::setupDimensions( dimensions );
  delete cache;
  cache = new beliefCacheClass( "triangulationCache", 
				cacheSize,
				clusterSize, 
				beliefConstructor,
				(void* (*)()) &triangulationCachedDataClass::constructor,
				&triangulationCachedDataClass::destructor );
}

void
triangulationClass::finalize( void )
{
  delete[] lambda;
  delete[] base;
  delete[] delta;
  delete[] x;
  delete simplexCorner;
  delete cache;

  if( vec != NULL )
    for( int i = 0; i < dimensions + 1; ++i )
      delete[] vec[i];
  delete[] vec;

  lambda = NULL;
  base = NULL;
  delta = NULL;
  x = NULL;
  vec = NULL;
  cache = NULL;
}

void
triangulationClass::clean( void )
{
  if( cache )
    cache->clean();
}

void
triangulationClass::triangulate( beliefHashClass* beliefHash,
				 bool useCache, 
				 const standardBeliefClass* belief, 
				 hashEntryClass**& simplex,
				 map<int,float>*& barycentricCoordinates )
{
  register int i, j;
  register float *basePtr;
  register hashEntryClass *entry;
  triangulationCachedDataClass *cachedSimplex;
  const map<int,float> *m;
  map<int,float>::const_iterator it;

  // cache lookup
  if( useCache && ((cachedSimplex = (triangulationCachedDataClass*)cache->lookup( belief, 0 )) != NULL) )
    {
      // restore barycentricCoordinates and simplex. O(1).
      goto remap;
    }
  else
    {
      // evict someone from cache
      cachedSimplex = (triangulationCachedDataClass*)cache->freeSlot( belief, 0 );
      basePtr = base;
    }

  // compute X = invMatrixB times bel. Worst-case O(n^2).
  m = belief->cast();
  memset( x, 0, dimensions * sizeof( float ) );
  for( it = m->begin(); it != m->end(); ++it )
    for( i = 0; i < dimensions; ++i )
      x[i] += ((*it).first >= i ? (*it).second : 0.0);
  for( i = 0; i < dimensions; ++i )
    x[i] = x[i] * levels;

  // find simplex S = (v^1, v^2, ..., v^n+1) containing X. O(n).
  for( i = 0; i < dimensions; ++i )
    {
      basePtr[i] = (float) floor( (double) (x[i] + PD.epsilon) );
      delta[i].idx = i;
      delta[i].val = x[i] - basePtr[i];
    }

  // sorting delta's. O(nlog(n)).
  qsort( delta, dimensions, sizeof( pair_s ), &pairCompare );

  // compute barycentric coordinates of X inside S. O(n).
  //xxxxxx OLD lambda[dimensions] = delta[dimensions-1].val;
  lambda[dimensions-1] = delta[dimensions-1].val;
  lambda[0] = 1.0 - lambda[dimensions];
  for( i = dimensions-1; i > 0; --i )
    {
      lambda[i] = delta[i-1].val - delta[i].val;
      lambda[0] -= lambda[i];
    }

  fill( cachedSimplex->barycentricCoordinates, lambda, dimensions + 1 );

  // compute simplex corners. Worst-case O(n^2).
  for( it = cachedSimplex->barycentricCoordinates.begin(); it != cachedSimplex->barycentricCoordinates.end(); ++it )
    {
      j = (*it).first;
      memcpy( vec[j], basePtr, dimensions * sizeof( float ) );
      for( i = j - 1; i >= 0; --i )
	vec[j][delta[i].idx] += 1.0;
    }

  // remap S into [0,1]^n. Worst-case O(n^2).
  for( it = cachedSimplex->barycentricCoordinates.begin(); it != cachedSimplex->barycentricCoordinates.end(); ++it )
    {
      j = (*it).first;
      for( i = 0; i < dimensions-1; ++i )
	x[i] = (vec[j][i] - vec[j][i+1]) / levels;
      x[dimensions-1] = vec[j][dimensions-1] / levels;
      simplexCorner->fill( x, dimensions );

      // populate belief hash
      if( (entry = beliefHash->lookup( (void*)simplexCorner )) == NULL )
	entry = beliefHash->insert( (void*)simplexCorner, simplexCorner->heuristicValue(), (void*)false );
      cachedSimplex->simplex[j] = entry;
    }

  // set return data. O(1).
 remap:
  barycentricCoordinates = &cachedSimplex->barycentricCoordinates;
  simplex = cachedSimplex->simplex;
}

static int
pairCompare( const void* v1, const void* v2 )
{
  pair_s *p1, *p2;
  float  delta;

  p1 = (pair_s*) v1;
  p2 = (pair_s*) v2;

  delta = p2->val - p1->val;
  if( delta > 0.0 )
    return( 1 );
  else if( delta < 0.0 )
    return( -1 );
  else
    return( 0 );
}
